# -*- coding: utf-8 -*-

from django import template
from django.utils.safestring import mark_safe

# We need a registration instance in order to configure everything with Django
register = template.Library()


@register.filter(name="chemical_formula")
def formula_to_html(formula_str):
    """
    Converts a chemical formula to html format by wrapping numbers with a
    subscript tag. For example, "Y2CF2" will be turned into "Y<sub>2</sub>CF<sub>2</sub>"
    so that it looks nice in a webpage
    """

    # start with an empty string that we build off of
    new_formula_str = ""

    for character in formula_str:
        # if it's a number or a decimal, we wrap it in <sub> tags
        if character.isnumeric() or character == ".":
            character = f"<sub>{character}</sub>"
        # now add it to our result
        new_formula_str += character

    # Because we added new html to our script, we need to have Django check it
    # ensure it safe before returning. Read more about this here:
    # https://docs.djangoproject.com/en/3.2/howto/custom-template-tags/#filters-and-auto-escaping
    return mark_safe(new_formula_str)



# -*- coding: utf-8 -*-

import os
import shutil
import yaml
import numpy

from pymatgen.analysis.transition_state import NEBAnalysis

from simmate.toolkit.diffusion import MigrationImages
from simmate.calculators.vasp.inputs import Incar, Poscar, Potcar
from simmate.calculators.vasp.tasks.base import VaspTask


class VaspNudgedElasticBandTask(VaspTask):
    """
    A base class for Nudged Elastic Band (NEB) calculations. This is not meant
    to be used directly but instead should be inherited from.

    Note, NEB tasks are very different from all other VASP tasks!

    The first big difference is that it takes a list of structures instead of
    just one structure. This means that instead of "structure=...", you should
    actually do "structures=[structure1, structure2, structure3, ...]", where
    the list of structures is your list of images.

    The second big difference is that VASP uses a different folder setup when
    running these calculations. It has a series of folders named 00, 01, 02, ... N,
    where 00 is the starting image, N is the endpoint image, and 01 to (N-1) are
    the midpoint images. Simmate handles this inside the task, but knowing this
    may be useful if you'd like to make your own variation of this class.
    """

    # NEB does not require a POSCAR file because input structures are organized
    # into folders.
    required_files = ["INCAR", "POTCAR"]

    # Pymatgen's NEB parser does not read from the vasprun.xml so it can't
    # confirm convergence here. I'll have to write my own output class to do this.
    confirm_convergence = False

    @classmethod
    def _pre_checks(
        cls,
        migration_images: MigrationImages,
        directory: str,
    ):
        """
        Runs a series of checks to ensure the user configured the job correctly.

        This is called automatically within the setup() method and shouldn't be
        used directly.
        """

        # The next common mistake is to mislabel the number of images in the
        # INCAR file.
        # first, we check if the user set this.
        nimages = cls.incar.get("IMAGES")
        if nimages:
            # if so, we check that it was set correctly. It should be equal to
            # the number of structures minus 2 (because we don't count the
            # start and end images here.)
            if nimages != (len(migration_images) - 2):
                raise Exception(
                    "IMAGES looks to be improperly set! This value should not"
                    " include the start/end images -- so make sure you counted"
                    " properly. Alternatively, you also can remove this keyword"
                    " from your INCAR and Simmate will provide it automatically"
                    " for you."
                )

        # TODO: add a precheck that ensures the number of cores VASP is ran on
        # is also divisible by the number of images. For example...
        # "mpirun -n 16 vasp" will not work for IMAGES=3 because 16 is not
        # divisible by 3. But this also may be better suited for an ErrorHandler.
        # An example error message from from VASP is...
        #       "M_divide: can not subdivide 16 nodes by 3"

        # make sure all images are contained with the cell
        migration_images_cleaned = cls._process_structures(migration_images)
        return migration_images_cleaned

    @staticmethod
    def _process_structures(structures: MigrationImages):
        """
        Remove any atom jumps across the cell.

        This method is copied directly from pymatgen's MITNEBset and has not
        been refactored/reviewed yet.

        TODO: This code would be better placed as a method of MigrationImages
        """
        input_structures = structures
        structures = [input_structures[0]]
        for s in input_structures[1:]:
            prev = structures[-1]
            for i, site in enumerate(s):
                t = numpy.round(prev[i].frac_coords - site.frac_coords)
                if numpy.any(numpy.abs(t) > 0.5):
                    s.translate_sites([i], t, to_unit_cell=False)
            structures.append(s)
        return MigrationImages(structures)  # convert back to simmate object

    @classmethod
    def setup(
        cls,
        directory: str,
        migration_images: MigrationImages,
        **kwargs,
    ):
        """
        Writes input files for a NEB calculation. Each structure image recieves
        it's own folder within the parent directory.

        This method is typically not called directly. Instead, users should
        use the `run` method which calls setup within it.

        #### Parameters

        - `structure`:
            This parameter does NOTHING! NEB is a special-case workflow that
            accepts a list of structures instead of a single one. Therefore, it
            is strictly for compatibility with the core S3Task. Leave this
            value at None.

        - `directory`:
            The name of the directory to write all input files in. This directory
            should exists before calling. (see utilities.get_directory)

        - `structures`:
            The list of structures to use as a MigrationImages object.
        """
        # !!! The structure input is confusing for users, so I should consider
        # removing it from the S3Task...

        # run some prechecks to make sure the user has everything set up properly.
        migration_images_cleaned = cls._pre_checks(migration_images, directory)

        # Here, each image (start to end structures) is put inside of its own
        # folder. We make those folders here, where they are named 00, 01, 02...N
        # Also recall that "structure" is really a list of structures here.
        for i, image in enumerate(migration_images_cleaned):

            # first make establish the foldername
            # The zfill function converts numbers from "1" to "01" for us
            foldername = os.path.join(directory, str(i).zfill(2))
            # see if the folder exists, and if not, make it
            if not os.path.exists(foldername):
                os.mkdir(foldername)
            # now write the poscar file inside the folder
            Poscar.to_file(image, os.path.join(foldername, "POSCAR"))

        # We also need to check if the user set IMAGES in the INCAR. If not,
        # we set that for them here. Note, we use the "pop" method which removes
        # the IMAGES__auto while grabbing its value. Because we are modifying
        # the incar dictionary here, we must make a copy of it -- this ensures
        # no bugs when this task is called in parallel.
        incar = cls.incar.copy()
        # !!! Should this code be moved to the INCAR class? Or would that require
        # too much reworking to allow INCAR to accept a list of structures?
        if not incar.get("IMAGES") and incar.pop("IMAGES__auto", None):
            incar["IMAGES"] = len(migration_images_cleaned) - 2

        # Combine our base incar settings with those of our parallel settings
        # and then write the incar file
        incar = Incar(**incar) + Incar(**cls.incar_parallel_settings)
        incar.to_file(
            filename=os.path.join(directory, "INCAR"),
            # we can use the start image for our structure -- as all structures
            # should give the same result.
            structure=migration_images_cleaned[0],
        )

        # if KSPACING is not provided in the incar AND kpoints is attached to this
        # class instance, then we write the KPOINTS file
        if cls.kpoints and ("KSPACING" not in cls.incar):
            raise Exception(
                "Custom KPOINTS are not supported by Simmate yet. "
                "Please use KSPACING in your INCAR instead."
            )

        # write the POTCAR file in this folder as well. Only one is needed and
        # all images will use refer to this POTCAR in the base directory
        Potcar.to_file_from_type(
            # we can use the start image for our structure -- as all structures
            # should give the same result.
            migration_images_cleaned[0].composition.elements,
            cls.functional,
            os.path.join(directory, "POTCAR"),
            cls.potcar_mappings,
        )

        # For the user's reference, we also like to write an image of the
        # starting path to a cif file. This can be slow for large structures
        # (>1s), but it is very little time compared to a full NEB run.
        path_vis = migration_images_cleaned.get_sum_structure()
        path_vis.to("cif", os.path.join(directory, "path_relaxed_idpp.cif"))

    @classmethod
    def setup_restart(cls, directory: str, **kwargs):
        """
        From a working directory of a past calculation, sets up for the calculation
        to be restarted.
        """
        print("WARNING: CONTCARs are not yet copied to POSCARs for NEB restarts.")

        # establish filenames
        stopcar_filename = os.path.join(directory, "STOPCAR")

        # delete the stopcar if it exists
        if os.path.exists(stopcar_filename):
            os.remove(stopcar_filename)

    @classmethod
    def workup(cls, directory: str):
        """
        Works up data from a NEB run, including confirming convergence and
        writing summary output files (structures, data, and plots).

        #### Parameters

        - `directory`:
            Name of the base folder where all results are located.
        """
        # BUG: For now I assume there are start/end image directories are located
        # in the working directory. These relaxation are actually ran by a
        # separate workflow, which is thus a prerequisite for this workflow.
        start_dirname = os.path.join(directory, "relaxation.vasp.neb-endpoint.start")
        end_dirname = os.path.join(directory, "relaxation.vasp.neb-endpoint.end")
        try:
            assert os.path.exists(start_dirname)
            assert os.path.exists(end_dirname)
        except:
            raise Exception(
                "Your NEB calculation completed successfully. However, in order "
                "to run the workup, Simmate needs the start/end point relaxations. "
                "These should be in the same folder as your NEB run and named "
                "relaxation.vasp.neb-endpoint.start and relaxation.vasp.neb-endpoint.end."
            )

        ################
        # BUG: NEBAnalysis.from_dir is broken for all folder structures except
        # when the start/end points are in the 00 and N folders. I therefore
        # need to copy the OUTCAR from the endpoint relaxations to these folders.
        # I don't want to mess with opening a pull request with them / waiting
        # on a new release, so I make this hacky fix here
        new_start_filename = os.path.join(directory, "00", "OUTCAR")

        # the end filename should be the highest number in the directory
        numbered_dirs = [d for d in os.listdir(directory) if d.isdigit()]
        numbered_dirs.sort()
        new_end_filename = os.path.join(directory, numbered_dirs[-1], "OUTCAR")

        # now copy the outcars over
        shutil.copyfile(os.path.join(start_dirname, "OUTCAR"), new_start_filename)
        shutil.copyfile(os.path.join(end_dirname, "OUTCAR"), new_end_filename)
        ################

        neb_results = NEBAnalysis.from_dir(
            directory,
            # BUG: see bug fix right above this
            # relaxation_dirs=[
            #     "endpoint_relaxation_start",
            #     "endpoint_relaxation_end",
            # ],
        )

        # write output files/plots for the user to quickly reference
        cls._write_output_summary(directory, neb_results)

        return neb_results

    @staticmethod
    def _write_output_summary(directory: str, neb_results: NEBAnalysis):
        """
        This is an EXPERIMENTAL feature.

        This prints a "simmate_summary.yaml" file with key output information.

        This method should not be called directly as it used within workup().
        """

        # plot the results
        # plot = neb_results.get_plot()
        # plot.savefig(os.path.join(directory, "NEB_plot.jpeg"))
        print(
            "NEB plot generation is temporarily disabled due to a bug in prefect "
            "where matplotlib.plot() leads to a segmentation fault. "
            "See https://github.com/PrefectHQ/prefect/issues/5991"
        )

        # convert all the structures to a MigrationImages object so we can write
        # the summed structure.
        migration_images = MigrationImages(neb_results.structures)
        structure_sum = migration_images.get_sum_structure()
        structure_sum.to("cif", os.path.join(directory, "path_relaxed_neb.cif"))

        results_dict = neb_results.as_dict()
        summary = {
            "structures": "Final structure images are the CONTCARs within image directories (00-N)",
            "forces_tangent": results_dict["forces"],
            "energies": results_dict["energies"],
            "structure_distances": results_dict["r"],
            "energy_barrier": float(
                max(neb_results.energies) - min(neb_results.energies)
            ),
        }

        summary_filename = os.path.join(directory, "simmate_summary.yaml")
        with open(summary_filename, "w") as file:
            content = yaml.dump(summary)
            file.write(content)



# -*- coding: utf-8 -*-

import os
import shutil

from pymatgen.io.vasp.outputs import Vasprun

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Unconverged(ErrorHandler):
    """
    Check if a calculation converged successfully. If not, the fix depends on
    whether its the ionic steps or electronic steps are struggling to converge.

    NOTE: This can be confusing when used along with the Nonconverging error handler
    because both attempt to fix failures in electronic convergence. These handlers
    are converted directly from Custodian, so it's not clear to me why they
    separated them. In the future, I need to clean this up.
    """

    is_monitor = False

    def check(self, directory: str) -> bool:

        # Now check if the calculation converged. If not, we have the error!
        xml_filename = os.path.join(directory, "vasprun.xml")
        try:
            # load the xml file and only parse the bare minimum
            xmlReader = Vasprun(
                filename=xml_filename,
                parse_dos=False,
                parse_eigen=False,
                parse_projected_eigen=False,
                parse_potcar_file=False,
                exception_on_bad_xml=True,
            )
            if not xmlReader.converged:
                return True
        except Exception:
            pass

        # if the xml fails to load or if it converged successfully, then we
        # don't have this error.
        return False

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # Also load the vasp results to see if it's the electronic or ionic steps
        # that are failing to converge.
        xml_filename = os.path.join(directory, "vasprun.xml")
        xmlReader = Vasprun(
            filename=xml_filename,
            parse_dos=False,
            parse_eigen=False,
            parse_projected_eigen=False,
            parse_potcar_file=False,
            exception_on_bad_xml=True,
        )

        # check if the electronic steps converged
        if not xmlReader.converged_electronic:
            # Here we switch the ALGO from VeryFast --> Fast --> Normal as
            # each ALGO is increasingly more stable but computationally more
            # expensive.

            # grab what was previously used as the ALGO, where the default is Normal
            current_algo = incar.get("ALGO", "Normal")

            # If the current algo is VeryFast, then switch it to Fast
            if current_algo == "VeryFast":
                # Set the new value
                incar["ALGO"] = "Fast"
                # rewrite the new INCAR file
                incar.to_file(incar_filename)
                # return the description of what we did
                return "switched ALGO from VeryFast to Fast"

            # If the current algo is Fast, then switch it to Normal
            elif current_algo == "Fast":
                # Set the new value
                incar["ALGO"] = "Normal"
                # rewrite the new INCAR file
                incar.to_file(incar_filename)
                # return the description of what we did
                return "switched ALGO from Fast to Normal"

            # If the current algo is Normal, then switch it to All
            elif current_algo == "Normal":
                # Set the new value
                incar["ALGO"] = "All"
                # rewrite the new INCAR file
                incar.to_file(incar_filename)
                # return the description of what we did
                return "switched ALGO from Normal to All"

            # As a last resort we try to introduce mixing
            else:
                new_settings = {
                    "ISTART": 1,
                    "ALGO": "Normal",
                    "NELMDL": -6,
                    "BMIX": 0.001,
                    "AMIX_MAG": 0.8,
                    "BMIX_MAG": 0.001,
                }
                # check that we dont already have these values set first
                if not all(
                    incar.get(key) == value for key, value in new_settings.items()
                ):
                    # Set the new values
                    incar.update(new_settings)
                    # rewrite the new INCAR file
                    incar.to_file(incar_filename)
                    # return the description of what we did
                    return (
                        f"turned on mixing with the following settings: {new_settings}"
                    )

        # if it's not electronic convergence, check it is the ionic steps
        elif not xmlReader.converged_ionic:
            # We continue the optimization with RMM-DIIS type relaxation

            # Copy the CONTCAR into the POSCAR
            poscar_filename = os.path.join(directory, "POSCAR")
            contcar_filename = os.path.join(directory, "CONTCAR")
            shutil.copy(contcar_filename, poscar_filename)
            correction = "copied the CONTCAR into the POSCAR"

            # update the INCAR if it isn't already IBRION=1
            current_ibrion = incar.get("IBRION", 0)
            if current_ibrion != 1:
                incar["IBRION"] = 1
                incar.to_file(incar_filename)
                correction += " and switched IBRION to 1"

            return correction



# -*- coding: utf-8 -*-

"""

This file is for pulling MPDS data into the Simmate database. 

MPDS provides a python package to make downloading their data very easy. This 
officially supported and maintained at https://github.com/mpds-io/mpds_client.
For now the package is only available via a pip install. And to help with using
this package, you should read through their guide and examples here:    
https://mpds.io/developer/#Client-library

NOTE: I'm currently running into permission issues with MPDS, and will return to
this provider after talking with their team.

"""

from django.db import transaction

from tqdm import tqdm
from simmate.toolkit import Structure
from mpds_client import MPDSDataRetrieval, MPDSDataTypes

from simmate.database import connect

from simmate.database.third_parties.mpds import MpdsStructure


@transaction.atomic
def load_all_structures(api_key="9WPgw8xYqYcB91ylff1CogIaXd4MGq6RoEZjwC1L0fh66uMw"):

    # !!! Remove key in production

    # connect to the database
    client = MPDSDataRetrieval(api_key)

    # By default, only structures from peer-reviewed papers are shown, but we
    # can switch this to include MPDS's in-house machine learning and first
    # principles structures too.
    # !!! only the ab initio and machine learning datasets are free and open
    # so I have to limit myself to those for now.
    #   client.dtype = MPDSDataTypes.ALL
    client.dtype = MPDSDataTypes.AB_INITIO

    # get_data returns a list, while get_dataframe will give us a pandas dataframe
    # I just use a list for now.
    # !!! This doesn't behave the way they claim... it definitely doesn't return
    # JSON. I'm also only ever getting hits for P-type ids (properties), when I
    # want S-type (structures)
    data = client.get_data(
        search={
            "elements": "Al-N",  # !!! for testing
            # "props": "atomic structure",
        },
        # fields={
        #     "S": [
        #         # "phase_id",
        #         # "entry",
        #         'chemical_formula',
        #         # "cell_abc",
        #         # "sg_n",
        #         # "basis_noneq",
        #         # "els_noneq",
        #     ]
        # },
    )



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.base import VaspTask
from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS,
)
from simmate.calculators.vasp.error_handlers import (
    TetrahedronMesh,
    RotationMatrix,
    Brmix,
    Zpotrf,
    SubspaceMatrix,
    IncorrectShift,
    RealOptlay,
    Tetirr,
    RotationNonIntMatrix,
    LongVector,
    TripleProduct,
    Pricel,
    Brions,
    Zbrent,
    InsufficientBands,
    Pssyevx,
    Eddrmm,
    Edddav,
    Edwav,
    Zheev,
    ElfKpar,
    Rhosyg,
    Posmap,
    PointGroup,
    SymprecNoise,
    IncorrectSmearing,
    MeshSymmetry,
    Unconverged,
    NonConverging,
    Potim,
    PositiveEnergy,
    Frozen,
    LargeSigma,
    Walltime,
)


class MatprojRelaxation(VaspTask):
    """
    This task is a reimplementation of pymatgen's
    [MPRelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPRelaxSet).

    Runs a VASP geometry optimization using Materials Project settings.

    Materials Project settings are often considered the minimum-required
    quality for publication and is sufficient for most applications. If you are
    looking at one structure in detail (for electronic, vibrational, and other
    properties), you should still test for convergence using higher-quality
    settings.
    """

    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS

    incar = dict(
        ALGO="Fast",
        EDIFF__per_atom=5.0e-05,
        ENCUT=520,
        IBRION=2,
        ISIF=3,
        ISMEAR=-5,
        ISPIN=2,
        KSPACING=0.4,  # !!! This is where we are different from pymatgen right now
        LASPH=True,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        NELM=100,
        NSW=99,
        PREC="Accurate",
        SIGMA=0.05,
        MAGMOM__smart_magmom={
            "default": 0.6,  # note this is different from the VASP default
            "Ce": 5,
            "Ce3+": 1,
            "Co": 0.6,
            "Co3+": 0.6,
            "Co4+": 1,
            "Cr": 5,
            "Dy3+": 5,
            "Er3+": 3,
            "Eu": 10,
            "Eu2+": 7,
            "Eu3+": 6,
            "Fe": 5,
            "Gd3+": 7,
            "Ho3+": 4,
            "La3+": 0.6,
            "Lu3+": 0.6,
            "Mn": 5,
            "Mn3+": 4,
            "Mn4+": 3,
            "Mo": 5,
            "Nd3+": 3,
            "Ni": 5,
            "Pm3+": 4,
            "Pr3+": 2,
            "Sm3+": 5,
            "Tb3+": 6,
            "Tm3+": 2,
            "V": 5,
            "W": 5,
            "Yb3+": 1,
        },
        LMAXMIX__smart_lmaxmix=True,
        multiple_keywords__smart_ldau=dict(
            LDAU__auto=True,
            LDAUTYPE=2,
            LDAUPRINT=1,
            LDAUJ={},  # pymatgen sets these, but they're all default values anyways
            LDAUL={
                "F": {
                    "Co": 2,
                    "Cr": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Ni": 2,
                    "V": 2,
                    "W": 2,
                },
                "O": {
                    "Co": 2,
                    "Cr": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Ni": 2,
                    "V": 2,
                    "W": 2,
                },
            },
            LDAUU={
                "F": {
                    "Co": 3.32,
                    "Cr": 3.7,
                    "Fe": 5.3,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Ni": 6.2,
                    "V": 3.25,
                    "W": 6.2,
                },
                "O": {
                    "Co": 3.32,
                    "Cr": 3.7,
                    "Fe": 5.3,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Ni": 6.2,
                    "V": 3.25,
                    "W": 6.2,
                },
            },
        ),
    )

    error_handlers = [
        TetrahedronMesh(),
        RotationMatrix(),
        Brmix(),
        Zpotrf(),
        SubspaceMatrix(),
        IncorrectShift(),
        RealOptlay(),
        Tetirr(),
        RotationNonIntMatrix(),
        LongVector(),
        TripleProduct(),
        Pricel(),
        Brions(),
        Zbrent(),
        InsufficientBands(),
        Pssyevx(),
        Eddrmm(),
        Edddav(),
        Edwav(),
        Zheev(),
        ElfKpar(),
        Rhosyg(),
        Posmap(),
        PointGroup(),
        SymprecNoise(),
        IncorrectSmearing(),
        MeshSymmetry(),
        Unconverged(),
        NonConverging(),
        Potim(),
        PositiveEnergy(),
        Frozen(),
        LargeSigma(),
        Walltime(),
    ]



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.relaxation.quality_04 import Quality04Relaxation


class Quality04Energy(Quality04Relaxation):
    """
    Runs a rough VASP static energy calculation.

    `Quality 04` relates to our ranking of relaxation qualities, where this
    calculation uses the same settings as the Quality04Relaxation.

    Note, even though this is currently our highest quality preset, these
    settings are still only suitable for high-throughput calculations or massive
    supercells. Settings are still below MIT and Materials Project quality.

    Most commonly, this is used in evolutionary searches (for structure
    prediction). We recommend instead using the relaxation/staged workflow,
    which uses this calculation as the sixth and final step -- after a series
    of rough relaxations are done.
    """

    # The settings used for this calculation are based on the MITRelaxation, but
    # we are updating/adding new settings here.
    # !!! we hardcode temperatures and time steps here, but may take these as inputs
    # in the future
    incar = Quality04Relaxation.incar.copy()
    incar.update(
        dict(
            ALGO="Normal",
            IBRION=-1,  # (optional) locks everything between ionic steps
            NSW=0,  # this is the main static energy setting
            ISMEAR=-5,  # was 0 for non-metals and 1 for metals
            SIGMA=0.05,  # was 0.05 for non-metals and 0.06 for metals
        )
    )
    # We set ISMEAR=0 and SIGMA above, so we no longer need smart_ismear
    incar.pop("multiple_keywords__smart_ismear")



# -*- coding: utf-8 -*-

import pandas


def ACF(filename="ACF.dat"):

    # open the file, grab the lines, and then close it
    with open(filename) as file:
        lines = file.readlines()

    # establish the headers. Note that I ignore the '#' column as this is
    # just site index.
    headers = ("x", "y", "z", "charge", "min_dist", "atomic_vol")

    # create a list that we will load data into
    bader_data = []
    # The first 2 lines are header and the final 4 lines are the footer. This is always
    # true so we don't need to iterate through those. The data we want is between the
    # header and footer so that's what we loop through.
    for line in lines[2:-4]:
        # By running strip, we convert the line from a string to a list of
        # The values are all still strings, so we convert them to int/floats
        # before saving. I add [1:] because the first value is just '#' which
        # is site index and we dont need.
        line_data = [eval(value) for value in line.split()[1:]]
        # add the line data to our ouput
        bader_data.append(line_data)

    # convert the list to a pandas dataframe
    dataframe = pandas.DataFrame(
        data=bader_data,
        columns=headers,
    )

    # Extra data is included in the footer that we can grab too. For each line, the data
    # is a float that is at the end of the line, hence the split()[-1].
    extra_data = {
        "vacuum_charge": float(lines[-3].split()[-1]),
        "vacuum_volume": float(lines[-2].split()[-1]),
        "nelectrons": float(lines[-1].split()[-1]),
    }

    return dataframe, extra_data



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate workflow engine. All commands are 
accessible through the "simmate workflow-engine" command.
"""

import click


@click.group()
def workflow_engine():
    """
    A group of commands for starting up Prefect Agents and Dask Clusters.
    There is also a simple "Simmate Cluster" that you can use too, but it
    is only meant for quick testing.

    Setting up your computational resources can be tricky, so be sure to go
    through our tutorial before trying this on your own: <TODO: INSERT LINK>
    """
    pass


@workflow_engine.command()
@click.option(
    "--queue_name",
    "-q",
    default=None,
    help="the unique name to give the work queue",
)
@click.option(
    "--concurrency_limit",
    "-c",
    default=1,
    help="the max number of workflow runs to run in parallel",
)
@click.option(
    "--nflows_max",
    "-n",
    default=None,
    help="the number of workflows runs to submit before shutdown",
)
@click.option(
    "--timeout",
    "-t",
    default=None,
    help="the time (in seconds) after which this worker will stop running jobs and shutdown",
)
@click.option(
    "--close_on_empty_queue",
    "-e",
    default=False,
    help="whether to shutdown when the queue is empty",
)
def start_worker(
    queue_name,
    concurrency_limit,
    nflows_max,
    timeout,
    close_on_empty_queue,
):
    """
    This starts a Simmate Worker which will query the database for jobs to run.
    """

    from simmate.workflow_engine import Worker

    worker = Worker(
        queue_name=queue_name,
        concurrency_limit=concurrency_limit,
        nflows_max=nflows_max,
        timeout=timeout,
        close_on_empty_queue=close_on_empty_queue,
    )
    worker.run()


@workflow_engine.command()
def start_singleflow_worker():
    """
    This starts a Simmate Worker that only runs one job and then shuts down. Also,
    if no job is available in the queue, it will shut down.

    Note: this can be acheived using the start-worker command too, but because
    this is such a common use-case, we include this command for convienence.
    It is the same as...

    simmate workflow-engine start-worker -c 1 -n 1 -e True
    """

    from simmate.workflow_engine import Worker

    worker = Worker(
        concurrency_limit=1,
        nflows_max=1,
        close_on_empty_queue=True,
    )
    worker.run()


@workflow_engine.command()
@click.option(
    "--agent_name",
    "-n",
    help="the name of the agent that will be visible in prefect cloud",
)
@click.option(
    "--agent_labels",
    "-l",
    help="labels that the agent will query for. To list multiple do... -l label1 -l label2",
    multiple=True,
)
@click.option(
    "--n_workers",
    "-w",
    help="the number of separate workers to run/submit",
    type=int,
)
@click.option(
    "--cpus_per_worker",
    "-c",
    help="the number of cpus that each worker will request",
    type=int,
)
@click.option(
    "--memory_per_worker",
    "-m",
    help="the amount of memory that each worker will request",
)
@click.option(
    "--walltime_per_worker",
    "-t",
    help="the timelimit set for each worker",
)
def run_cluster(
    agent_name,
    agent_labels,
    n_workers,
    cpus_per_worker,
    memory_per_worker,
    walltime_per_worker,
):
    """
    This starts up a Dask cluster and a Prefect Agent in order to run Simmate jobs.

    This convenience command is really only meant for basic purposes, as Dask
    and Prefect teams offer their own commands with much more control. For
    example, this command uses Prefect's LocalAgent, but there are other
    advanced types available such as DockerAgent which may be better for your
    use case.

    If you would like this cluster to run endlessly in the background, you can
    submit it with something like "nohup simmate workflow-engine run-cluster &".
    The "nohup" and "&" symbol together make it so this runs in the background AND
    it won't shutdown if you close your terminal (or ssh). To stop this from running,
    you'll now need to find the running process and kill it. Use something like
    "ps -aef | grep simmate" to find the running process and grab its ID. Then
    kill that process id with "kill 123" (if the id was 123).
    """
    raise NotImplementedError("This method is still being ported to Prefect v2")

    # All input arguments are optional. Therefore, I go through all of them and
    # remove the ones that weren't set. This prevents overwritting default settings
    # at a lower level. Note, I also rename several parameters (e.g. "cpus_per_worker").
    # These different names for the CLI only exist to make things easier for beginners.
    agent_kwargs = locals()  # grabs all input parameters as a dict
    possible_kwargs = list(agent_kwargs.keys())  # because we will be deleting keys
    for key in possible_kwargs:
        if not agent_kwargs[key]:
            agent_kwargs.pop(key)
            continue
        # renaming several parameters...
        if key == "cpus_per_worker":
            agent_kwargs["job_cpu"] = agent_kwargs.pop(key)
        elif key == "memory_per_worker":
            agent_kwargs["job_mem"] = agent_kwargs.pop(key)
        elif key == "walltime_per_job":
            agent_kwargs["walltime"] = agent_kwargs.pop(key)

    from simmate.configuration.prefect.setup_resources import (
        run_cluster_and_agent,
    )

    run_cluster_and_agent(**agent_kwargs)


# explicitly list functions so that pdoc doesn't skip them
__all__ = [
    "workflow_engine",
    "start_worker",
    "start_singleflow_worker",
    "run_cluster",
]



# -*- coding: utf-8 -*-

from pymatgen.analysis.structure_matcher import StructureMatcher

from simmate.toolkit import Structure
from simmate.workflow_engine import task, Workflow
from simmate.database.third_parties import MatprojStructure
from simmate.calculators.vasp.tasks.population_analysis import (
    MatprojPreBaderELF,
)
from simmate.calculators.bader.tasks import BaderELFAnalysis
from simmate.calculators.vasp.database.population_analysis import (
    MatprojBaderELFAnalysis as MPBadelfResults,
)


class PopulationAnalysis__Vasp__BadelfMatproj(Workflow):
    """
    Runs a static energy calculation using an extra-fine FFT grid and then
    carries out Bader analysis on the resulting charge density using the ELFCAR
    as a reference when partitioning.
    """

    database_table = MPBadelfResults

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        command: str = None,
        source: dict = None,
        directory: str = None,
    ):

        prebadelf_result = PopulationAnalysis__Vasp__PrebadelfMatproj.run(
            structure=structure,
            command=command,
            source=source,
            directory=directory,
        ).result()

        # load the structure that contains dummy atoms in it
        structure_w_empties = get_structure_w_empties(
            structure=structure,
            empty_ion_template="F",
        )

        # Bader only adds files and doesn't overwrite any, so I just run it
        # in the original directory. I may switch to copying over to a new
        # directory in the future though.
        badelf_result = BaderELFAnalysis.run(
            structure=structure_w_empties,
            directory=prebadelf_result["directory"],
        ).result()

        save_badelf_results(badelf_result, prebadelf_result["prefect_flow_run_id"])


# -----------------------------------------------------------------------------

# Below are extra tasks and subflows for the workflow that is defined above


class PopulationAnalysis__Vasp__PrebadelfMatproj(Workflow):
    s3task = MatprojPreBaderELF
    database_table = MPBadelfResults
    description_doc_short = "runs Bader analysis with ELFCAR as reference"


@task
def get_structure_w_empties(
    structure,
    empty_ion_template,
    # directory,
):
    """
    Searches the Materials Project database for a structure that contains
    the extra ion and matching host lattice, and uses it to introduce
    empty atoms into the original structure. For example, if your input
    gave a Ca2N structure and Cl ion for the template, this function will
    find Ca2NCl in the Matproj database and then use the Cl sites to
    add dummy "H" atoms into the Ca2N structure. This is particularly useful
    if you would like to analyze electrides.
    """

    # !!! It might make more sense to have this accept a database Structure object
    # and also search the same table for a match (rather than the Matproj).

    # Grab the chemical system of the structure when it includes the empty ion template
    structure = Structure.from_dynamic(
        structure
    )  # !!! because we don't load input in this flow yet
    structure_dummy = structure.copy()
    structure_dummy.append(
        empty_ion_template, [0, 0, 0]
    )  # coords don't matter here bc I'm just after the chemical_system
    template_system = structure_dummy.composition.chemical_system

    # Go through the Materials Project database and find a structure that
    # is matching when the empty ion type is ignored.
    matcher = StructureMatcher(
        ignored_species=empty_ion_template,
        primitive_cell=False,  # required for the get_s2_like_s1 method
    )
    potential_matches = MatprojStructure.objects.filter(
        chemical_system=template_system
    ).all()
    for potential_match in potential_matches:
        structure_template = potential_match.to_toolkit()
        is_match = matcher.fit(structure, structure_template)
        # Once we find a match, stop searching
        if is_match:
            break

    # confirm that we successfully found a match
    if not is_match:
        raise Exception("No match found when trying to generate empty sites")

    # convert our match to an equivalent basis set as the original input
    # structure -- this ensures frac coords are matches.
    structure_template = matcher.get_s2_like_s1(structure, structure_template)

    # make a copy because we will be modifying the structure
    structure_w_empties = structure.copy()

    for site in structure_template.sites:
        if site.specie.symbol == empty_ion_template:
            structure_w_empties.append("H", site.frac_coords)

    # TODO: consider doing a check to ensure the H atoms added are reasonable.
    # A simple check could be to use the SiteDistance validator and make sure
    # the hydrogen isn't too close to another atom.

    # write the new structure to file
    # filename = os.path.join(directory, "simmate_structure_w_empties.cif")
    # structure_w_empties.to("cif", filename)

    return structure_w_empties


# THIS IS A COPY/PASTE FROM THE BADER WORKFLOW -- I need to condense these
@task
def save_badelf_results(bader_result, prefect_flow_run_id):
    # load the results. We are particullary after the first result with
    # is a pandas dataframe of oxidation states.
    oxidation_data, extra_data = bader_result["result"]

    # load the calculation entry for this workflow run. This should already
    # exist thanks to the load_input_and_register task of the prebader workflow
    calculation = MPBadelfResults.from_prefect_id(
        prefect_flow_run_id,
    )
    # BUG: can't use context to grab the id because workflow tasks generate a
    # different id than the main workflow

    # now update the calculation entry with our results
    calculation.oxidation_states = list(oxidation_data.oxidation_state.values)
    calculation.save()



# -*- coding: utf-8 -*-

# For a list of parameter and their defaults look at...
#   https://docs.deepmodeling.org/projects/deepmd/en/master/train/train-input.html

{
    "model": {
        "descriptor": {
            "type": "se_e2_a",
            # "sel": [46, 92],  # Default is Auto
            "neuron": [25, 50, 100],  # Default is [2, 4, 8]
            "axis_neuron": 16,  # Default is 4
        },
        "fitting_net": {
            "neuron": [240, 240, 240],  # Default is [120, 120, 120]
            # "resnet_dt": True,  # Default is False
            # "trainable": True,  # Default is False
            # "atom_ener": [],  # Specify energy in vacuum for each type
        },
    },
    "learning_rate": {},
    "loss": {},
    "training": {
        "training_data": {
            "systems": [
                "../deepmd_data/C7_train",
                "../deepmd_data/C1_train",
                "../deepmd_data/C6_train",
                "../deepmd_data/C4_train",
                "../deepmd_data/C8_train",
                "../deepmd_data/C2_train",
                "../deepmd_data/C5_train",
                "../deepmd_data/C3_train",
                "../deepmd_data/C9_train",
            ]
        },
        "validation_data": {
            "systems": [
                "../deepmd_data/C4_test",
                "../deepmd_data/C8_test",
                "../deepmd_data/C6_test",
                "../deepmd_data/C2_test",
                "../deepmd_data/C5_test",
                "../deepmd_data/C9_test",
                "../deepmd_data/C3_test",
                "../deepmd_data/C7_test",
                "../deepmd_data/C1_test",
            ],
            # "batch_size": 1,  # Default is 1
            "numb_btch": 1,  # Default is 1
        },
        "numb_steps": 1000000,
        # "numb_test": 1,  # Default is 1
    },
}



# -*- coding: utf-8 -*-

"""
Calculate and plot both the density of states (DOS) and band structure
for a given crystal structure
"""

# TODO: when I add more calculators, I can do something like this...
# if "simmate.calculators.vasp" in installed_apps:

from simmate.calculators.vasp.workflows.electronic_structure.all import (
    ElectronicStructure__Vasp__MatprojFull,
)



# -*- coding: utf-8 -*-

from django_filters import rest_framework as filters

from simmate.database.base_data_types import DatabaseTable


class DatabaseTableFilter(filters.FilterSet):
    class Meta:
        table = DatabaseTable
        fields = {}

    @classmethod
    def get_mixins(cls):
        """
        Grabs the mix-in Forms that were used to make this class. This will
        be mix-ins like Structure, Forces, etc. from the
        `simmate.website.workflows.forms` module.
        """
        # We skip the first entry because it is always DatabaseTableForm
        return [parent_class for parent_class in cls.__bases__[1:]]

    @classmethod
    def get_mixin_names(cls):
        return [mixin.__name__ for mixin in cls.get_mixins()]

    @classmethod
    def get_extra_filters(cls):
        """
        Finds all columns that aren't covered by the supported Form mix-ins.

        For example, a form made from the database table...

        ``` python
        from simmate.database.base_data_types import (
            table_column,
            Structure,
            Forces,
        )

        class ExampleTable(Structure, Forces):
            custom_column1 = table_column.FloatField()
            custom_column2 = table_column.FloatField()
        ```

        ... would return ...

        ``` python
        ["custom_column1", "custom_column2"]
        ```
        """

        all_columns = cls.base_filters.keys()
        columns_w_mixin = [
            column for mixin in cls.get_mixins() for column in mixin.base_filters.keys()
        ]
        extra_columns = [
            column
            for column in all_columns
            if column not in columns_w_mixin and column != "id"
        ]
        return extra_columns

    @classmethod
    def from_table(cls, table: DatabaseTable):
        """
        Dynamically creates a Django Filter from a Simmate database table.

        For example, this function would take
        `simmate.database.third_parties.MatprojStructure`
        and automatically make the following filter:

        ``` python
        from simmate.website.core_components.filters import (
            DatabaseTableFilter,
            Structure,
            Thermodynamics,
        )


        class MatprojStrucureFilter(
            DatabaseTableFilter,
            Structure,
            Thermodynamics,
        ):
            class Meta:
                model = MatprojStructure  # this is database table
                fields = {...} # this combines the fields from Structure/Thermo mixins

            # These attributes are set using the declared filters from Structure/Thermo mixins
            declared_filter1 = ...
            declared_filter1 = ...
        ```
        """

        # this must be imported locally because it depends on all other classes
        # from this module -- and will create circular import issues if outside
        from simmate.website.core_components import filters as simmate_filters

        # First we need to grab the parent mixin classes of the table. For example,
        # the MatprojStructure uses the database mixins ['Structure', 'Thermodynamics']
        # while MatprojStaticEnergy uses ["StaticEnergy"].
        mixin_names = [base.__name__ for base in table.__bases__]

        # Because our Forms follow the same naming conventions as
        # simmate.database.base_data_types, we can simply use these mixin names to
        # load a Form mixin from the simmate.website.workflows.form module. We add
        # these mixins onto the standard ModelForm class from django.
        filter_mixins = [cls]
        filter_mixins += [
            getattr(simmate_filters, name)
            for name in mixin_names
            if hasattr(simmate_filters, name)
        ]

        # combine the fields of each filter mixin. Note we use .get_fields instead
        # of .fields to ensure we get a dictionary back
        filter_fields = {
            field: conditions
            for mixin in filter_mixins
            for field, conditions in mixin.get_fields().items()
        }

        # Also combine all declared filters from each mixin
        filters_declared = {
            name: filter_obj
            for mixin in filter_mixins
            for name, filter_obj in mixin.declared_filters.items()
        }

        # The FilterSet class requires that we set extra fields in the Meta class.
        # We define those here. Note, we accept all fields by default and the
        # excluded fields are only those defined by the supported form_mixins -
        # and we skip the first mixin (which is always DatabaseTableFilter)
        class Meta:
            model = table
            fields = filter_fields

        extra_attributes = {"Meta": Meta, **filters_declared}
        # BUG: __module__ may need to be set in the future, but we never import
        # these forms elsewhere, so there's no need to set it now.

        # Now we dynamically create a new form class that we can return.
        NewClass = type(table.__name__, tuple(filter_mixins), extra_attributes)

        return NewClass



# -*- coding: utf-8 -*-

import plotly.graph_objects as plotly_go
from django.apps import apps as django_apps
from prefect.client import get_client
from prefect.orion.schemas.filters import FlowRunFilter

from simmate.database.base_data_types import table_column, DatabaseTable
from simmate.utilities import async_to_sync


class EvolutionarySearch(DatabaseTable):
    """
    This database table holds all of the information related to an evolutionary
    search and also has convient methods to analyze the data.

    Loading Results
    ---------------

    Typically, you'll load your search through a search id or a composition:

    .. code-block:: python

        from simmate.shortcuts import SearchResults

        # if you know the id
        search_results = SearchResults.objects.get(id=123)

        # if you know the composition
        search_results = SearchResults.objects.get(id="Ca2 N1")

    Alternatively, you can find these out by looking at a table of all the
    evolutionary searches that have been ran:

    .. code-block:: python

        all_searches = SearchResults.objects.to_dataframe()

    Viewing Results
    ---------------

    The first thing you may want to check is the best structure found. To access
    this and write it to a file:

    .. code-block:: python

        # loads the best structure and converts it to a pymatgen structure object
        structure = search_results.best_individual.to_toolkit()

        # writes it to a cif file
        structure.to("cif", "best_structure.cif")

    To view convergence of the search, you can use the convenient plotting methods.

    Note: this will open up the plot in your default browser, so this command
    won't work properly through an ssh terminal.

    .. code-block:: python

        search_results.view_convergence_plot()

    If you are benchmarking Simmate to see if it found a particular structure,
    you can use:

    .. code-block:: python

        from simmate.toolkit import Structure

        structure = Structure.from_file("example123.cif")

        search_results.view_correctness_plot(structure)

    Beyond plots, you can also access a table of all calculated structures:

    .. code-block:: python

        dataframe = search_results.individuals_completed.to_dataframe()


    """

    class Meta:
        app_label = "workflows"

    # consider formula_full or chemical_system by making a composition-based mixin
    composition = table_column.CharField(max_length=50)  # !!! change to formula_full?

    # Import path for the workflow(s)
    individuals_datatable_str = table_column.CharField(max_length=200)

    # List of import paths for workflows used at any point. While all workflows
    # populate the individuals_datatable, they might do this in different ways.
    # For example, one may start with a ML prediction, another runs a series
    # of VASP relaxations, and another simply copies a structure from AFLOW.
    workflows = table_column.JSONField()

    # Import path that grabs the fitness value
    # I assume energy for now
    # fitness_function = table_column.CharField(max_length=200)

    max_structures = table_column.IntegerField()
    limit_best_survival = table_column.IntegerField()

    # relationships
    # sources / individuals
    # stop_conditions

    # get_stats:
    #   Total structure counts
    #   makeup of random, heredity, mutation, seeds, COPEX

    @property
    def individuals_datatable(self):
        # NOTE: this table just gives the class back and doesn't filter down
        # to the relevent individuals for this search. For that, use the "individuals"
        # property
        # we assume the table is registered in the local_calcs app
        return django_apps.get_model(
            app_label="workflows",
            model_name=self.individuals_datatable_str,
        )

    @property
    def individuals(self):
        # note we don't call "all()" on this queryset yet becuase this property
        # it often used as a "base" queryset (and additional filters are added)
        return self.individuals_datatable.objects.filter(formula_full=self.composition)

    @property
    def individuals_completed(self):
        # If there is an energy_per_atom, we can treat the calculation as completed
        return self.individuals.filter(energy_per_atom__isnull=False)
        # OPTIMIZE: would it be better to check energy_per_atom or structure_final?
        # Ideally, I could make a relation to the prefect flow run table but this
        # would require a large amount of work to implement.

    @property
    def best_individual(self):
        best = self.individuals_completed.order_by("energy_per_atom").first()
        return best

    def get_convergence_plot(self):

        # Grab the calculation's structure and convert it to a dataframe
        structures_dataframe = self.individuals_completed.to_dataframe()

        # There's only one plot here, no subplot. So we make the scatter
        # object and just pass it directly to a Figure object
        scatter = plotly_go.Scatter(
            x=structures_dataframe["updated_at"],
            y=structures_dataframe["energy_per_atom"],
            mode="markers",
        )
        figure = plotly_go.Figure(data=scatter)

        figure.update_layout(
            xaxis_title="Date Completed",
            yaxis_title="Energy (eV/atom)",
        )

        # we return the figure object for the user
        return figure

    def view_convergence_plot(self):
        figure = self.get_convergence_plot()
        figure.show(renderer="browser")

    def get_correctness_plot(self, structure_known):

        # --------------------------------------------------------
        # This code is from simmate.toolkit.validators.fingerprint.pcrystalnn
        # OPTIMIZE: There should be a convience method to make this featurizer
        # since I use it so much
        import numpy
        from simmate.toolkit import Composition
        from matminer.featurizers.site import CrystalNNFingerprint
        from matminer.featurizers.structure.sites import (
            SiteStatsFingerprint,  # PartialsSiteStatsFingerprint,
        )

        sitefingerprint_method = CrystalNNFingerprint.from_preset(
            "ops", distance_cutoffs=None, x_diff_weight=3
        )
        featurizer = SiteStatsFingerprint(
            sitefingerprint_method,
            stats=["mean", "std_dev", "minimum", "maximum"],
        )
        featurizer.elements_ = numpy.array(
            [element.symbol for element in Composition(self.composition).elements]
        )
        # --------------------------------------------------------

        # Grab the calculation's structure and convert it to a dataframe
        structures_dataframe = self.individuals_completed.to_dataframe()

        # because we are using the database model, we first want to convert to
        # pymatgen structures objects and add a column to the dataframe for these
        #
        #   structures_dataframe["structure"] = [
        #       structure.to_toolkit() for structure in ionic_step_structures
        #   ]
        #
        # BUG: the read_frame query creates a new query, so it may be a different
        # length from ionic_step_structures. For this reason, we can't iterate
        # through the queryset like in the commented out code above. Instead,
        # we need to iterate through the dataframe rows.
        # See https://github.com/chrisdev/django-pandas/issues/138 for issue
        from simmate.toolkit import Structure

        structures_dataframe["structure"] = [
            Structure.from_str(s.structure_string, fmt="POSCAR")
            for _, s in structures_dataframe.iterrows()
        ]

        from tqdm import tqdm

        structures_dataframe["fingerprint"] = [
            numpy.array(featurizer.featurize(s.structure))
            for _, s in tqdm(structures_dataframe.iterrows())
        ]

        fingerprint_known = numpy.array(featurizer.featurize(structure_known))

        structures_dataframe["fingerprint_distance"] = [
            numpy.linalg.norm(fingerprint_known - s.fingerprint)
            for _, s in tqdm(structures_dataframe.iterrows())
        ]

        # There's only one plot here, no subplot. So we make the scatter
        # object and just pass it directly to a Figure object
        scatter = plotly_go.Scatter(
            x=structures_dataframe["updated_at"],
            y=structures_dataframe["fingerprint_distance"],
            mode="markers",
            marker_color=structures_dataframe["energy_per_atom"],
        )
        figure = plotly_go.Figure(data=scatter)

        figure.update_layout(
            xaxis_title="Date Completed",
            yaxis_title="Distance from Known Structure",
        )

        # we return the figure object for the user
        return figure

    def view_correctness_plot(self, structure_known):
        figure = self.get_correctness_plot(structure_known)
        figure.show(renderer="browser")


class StructureSource(DatabaseTable):
    class Meta:
        app_label = "workflows"

    name = table_column.CharField(max_length=50)

    is_steadystate = table_column.BooleanField()
    is_singleshot = table_column.BooleanField()

    settings = table_column.JSONField(default=dict)

    # This list limits to ids that are submitted or running
    prefect_flow_run_ids = table_column.JSONField(default=list)

    search = table_column.ForeignKey(
        EvolutionarySearch,
        on_delete=table_column.CASCADE,
        related_name="sources",
    )

    @staticmethod
    @async_to_sync
    async def _check_still_running_ids(prefect_flow_run_ids):
        """
        Queries Prefect to see check on a list of flow run ids and determines
        which ones are still in a scheduled, running, or pending state.
        From the list of ids given, it will return a list of the ones that
        haven't finished yet.

        This is normally used within `update_flow_run_ids` and shouldn't
        be called directly.
        """

        # The reason we have this code as a separate method is because we want
        # to isolate Prefect's async calls from Django's sync-restricted calls
        # (i.e. django raises errors if called within an async context).

        async with get_client() as client:
            response = await client.read_flow_runs(
                flow_run_filter=FlowRunFilter(
                    id={"any_": prefect_flow_run_ids},
                    state={"type": {"any_": ["SCHEDULED", "PENDING", "RUNNING"]}},
                ),
            )

        still_running_ids = [str(entry.id) for entry in response]

        return still_running_ids

    def update_flow_run_ids(self):
        """
        Queries Prefect to see how many workflows are in a scheduled, running,
        or pending state from the list of run ids that are associate with this
        structure source.
        """

        # make the async call to Prefect client
        still_running_ids = self._check_still_running_ids(self.prefect_flow_run_ids)

        # we now have our new list of IDs! Let's update it to the database
        self.prefect_flow_run_ids = still_running_ids
        self.save()

        return still_running_ids

    @property
    def nprefect_flow_runs(self):
        # update our ids before we report how many there are.
        runs = self.update_flow_run_ids()
        # now the currently running ones is just the length of ids!
        return len(runs)



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import DatabaseTable, table_column

from pymatgen.symmetry.groups import SpaceGroup as PymatgenSpacegroup

# NOTE: This class is *not* meant to store data about each spacegroup, but instead
# is just to help with querying Structures. If you want all spacegroup data
# (such as the list of symmetry operations), then you should instead look at
# pymatgen.symmetry.groups.SpaceGroup. There is also spglib where you
# can get data using spglib.get_spacegroup_type(hall_number), but this is organized
# into 530 hall spacegroups, rather than the typical 230 international spacegroups.


class Spacegroup(DatabaseTable):
    class Meta:
        app_label = "core_components"

    source = None
    """
    The source column is disabled for this table because this is common
    information.
    """

    # NOTE: International spacegroup info is also known as Hermann-Mauguin

    number = table_column.IntegerField(primary_key=True)
    """
    International space group number
    """

    symbol = table_column.CharField(max_length=15)
    """
    Full international symbol
    """

    crystal_system = table_column.CharField(max_length=15)
    """
    The crystal system (don't confused with crystal family or lattice family)
    """

    point_group = table_column.CharField(max_length=15)
    """
    Point group symbol
    """

    @staticmethod
    def _load_database_from_toolkit():
        """
        Loads spacegroup data into the database table.

        This should never be called by the user. This method is automatically
        called when you first set up your database and should never be called
        after that.

        See `simmate.database.utilities.reset_database`
        """

        for number in range(1, 231):

            # load the pymatgen object that we can easily grab data from
            spacegroup = PymatgenSpacegroup.from_int_number(number)

            # load the data into this django model
            spacegroup_db = Spacegroup(
                number=spacegroup.int_number,
                symbol=spacegroup.symbol,
                crystal_system=spacegroup.crystal_system,
                point_group=spacegroup.point_group,
            )

            # now save this information to the database
            spacegroup_db.save()



# -*- coding: utf-8 -*-

from simmate.toolkit import Structure
from simmate.workflow_engine import task, Workflow
from simmate.calculators.vasp.tasks.population_analysis import (
    MatprojPreBader as MPPreBaderTask,
)
from simmate.calculators.vasp.database.population_analysis import (
    MatprojBaderAnalysis as MPBaderResults,
)
from simmate.calculators.bader.tasks import (
    BaderAnalysis as BaderAnalysisTask,
    CombineCHGCARs,
)


class PopulationAnalysis__Vasp__BaderMatproj(Workflow):
    """
    Runs a static energy calculation using an extra-fine FFT grid and then
    carries out Bader analysis on the resulting charge density.
    """

    database_table = MPBaderResults

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        command: str = None,
        source: dict = None,
        directory: str = None,
    ):

        prebader_result = PopulationAnalysis__Vasp__PrebaderMatproj.run(
            structure=structure,
            command=command,
            source=source,
            directory=directory,
        ).result()

        # Setup chargecars for the bader analysis and wait until complete
        CombineCHGCARs.run(directory=prebader_result["directory"]).result()

        # Bader only adds files and doesn't overwrite any, so I just run it
        # in the original directory. I may switch to copying over to a new
        # directory in the future though.
        bader_result = BaderAnalysisTask.run(
            directory=prebader_result["directory"]
        ).result()

        save_bader_results(bader_result, prebader_result["prefect_flow_run_id"])


# -----------------------------------------------------------------------------

# Below are extra tasks and subflows for the workflow that is defined above


class PopulationAnalysis__Vasp__PrebaderMatproj(Workflow):
    s3task = MPPreBaderTask
    database_table = MPBaderResults
    description_doc_short = "uses Materials Project settings with denser FFT grid"


@task
def save_bader_results(bader_result, prefect_flow_run_id):
    # load the results. We are particullary after the first result with
    # is a pandas dataframe of oxidation states.
    oxidation_data, extra_data = bader_result["result"]

    # load the calculation entry for this workflow run. This should already
    # exist thanks to the load_input_and_register task of the prebader workflow
    calculation = MPBaderResults.from_prefect_id(
        prefect_flow_run_id,
    )
    # BUG: can't use context to grab the id because workflow tasks generate a
    # different id than the main workflow

    # now update the calculation entry with our results
    calculation.oxidation_states = list(oxidation_data.oxidation_state.values)
    calculation.save()



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.static_energy import MatprojStaticEnergy


class MatprojNMRChemicalShifts(MatprojStaticEnergy):
    """
    This task is a reimplementation of pymatgen's
    [MPNMRSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPNonSCFSet)
    with mode="cs" (Chemical Shift).
    """

    incar = MatprojStaticEnergy.incar.copy()
    incar.update(
        dict(
            LCHIMAG=True,
            EDIFF=-1.0e-10,
            ISYM=0,
            LCHARG=False,
            LNMR_SYM_RED=True,
            NELMIN=10,
            NSLPLINE=True,
            PREC="ACCURATE",
            SIGMA=0.01,
        )
    )
    incar.pop("EDIFF__per_atom")



# -*- coding: utf-8 -*-

import pytest
from pytest_django.asserts import assertTemplateUsed
from django.urls import reverse


def test_providers_view(client):

    # grabs f"/third-parties/"
    url = reverse("third_parties:home")

    response = client.get(url)
    assert response.status_code == 200
    assertTemplateUsed(response, "third_parties/providers_all.html")


@pytest.mark.django_db
@pytest.mark.parametrize(
    "provider_name",
    [
        "AflowPrototype",
        "MatprojStructure",
        "CodStructure",
        "JarvisStructure",
        "OqmdStructure",
    ],
)
def test_workflows_by_type_view(client, provider_name):

    # list view

    # grabs f"/workflows/{workflow_type}/"
    url = reverse(
        "third_parties:provider",
        kwargs={"provider_name": provider_name},
    )

    response = client.get(url)
    assert response.status_code == 200
    assertTemplateUsed(response, "third_parties/provider.html")

    # detail view
    url = reverse(
        "third_parties:entry-detail",
        kwargs={
            "provider_name": provider_name,
            "pk": 999,
        },
    )

    response = client.get(url)
    assert response.status_code == 404
    # assertTemplateUsed(response, "third_parties/entry_detail.html")



# -*- coding: utf-8 -*-

import pytest

RADIUS_METHODS = [
    "ionic",
    "atomic",
    "atomic_calculated",
    "van_der_waals",
    # "metallic",  This is a special case that we test separately
    "ionic",
]


@pytest.mark.parametrize("radius_method", RADIUS_METHODS)
def test_radii_estimate(composition, radius_method):
    composition.radii_estimate(radius_method)


def test_radii_estimate_metallic(sample_compositions):

    # make sure a metallic composition succeeds
    composition_fe = sample_compositions["Fe1"]
    composition_fe.radii_estimate("metallic")

    # make sure a non-metallic composition fails
    with pytest.raises(Exception):
        composition_si = sample_compositions["Si2"]
        composition_si.radii_estimate("metallic")


@pytest.mark.parametrize("radius_method", RADIUS_METHODS)
def test_volume_estimate(composition, radius_method):
    composition.volume_estimate(radius_method)


@pytest.mark.parametrize("radius_method", RADIUS_METHODS)
def test_distance_matrix_estimate(composition, radius_method):
    composition.distance_matrix_estimate(radius_method)


def test_chemical_subsystems(sample_compositions):

    composition = sample_compositions["Fe1"]
    assert composition.chemical_subsystems == ["Fe"]

    composition = sample_compositions["Si2"]
    assert composition.chemical_subsystems == ["Si"]

    composition = sample_compositions["C4"]
    assert composition.chemical_subsystems == ["C"]

    composition = sample_compositions["Ti2O4"]
    assert composition.chemical_subsystems == ["O", "Ti", "O-Ti"]

    composition = sample_compositions["Al4O6"]
    assert composition.chemical_subsystems == ["Al", "O", "Al-O"]

    composition = sample_compositions["Si4N4O2"]
    assert composition.chemical_subsystems == [
        "N",
        "O",
        "Si",
        "N-O",
        "N-Si",
        "O-Si",
        "N-O-Si",
    ]

    composition = sample_compositions["Si4O8"]
    assert composition.chemical_subsystems == ["O", "Si", "O-Si"]

    composition = sample_compositions["Sr4Si4N8"]
    assert composition.chemical_subsystems == [
        "N",
        "Si",
        "Sr",
        "N-Si",
        "N-Sr",
        "Si-Sr",
        "N-Si-Sr",
    ]

    composition = sample_compositions["Mg4Si4O12"]
    assert composition.chemical_subsystems == [
        "Mg",
        "O",
        "Si",
        "Mg-O",
        "Mg-Si",
        "O-Si",
        "Mg-O-Si",
    ]



# -*- coding: utf-8 -*-

from selenium import webdriver
import pandas as pd
from tqdm import tqdm

# launch the webbrowser
driver = webdriver.Chrome(
    executable_path="/snap/bin/chromium.chromedriver"
)  # /snap/bin/chromium gives errors

# load the webpage
driver.get("http://aflowlib.org/CrystalDatabase/prototype_index.html")

table = driver.find_element_by_id("myTable")

rows = table.find_elements_by_tag_name("tr")

# grab all the headers from the first row
headers = [column.text for column in rows[0].find_elements_by_tag_name("th")]
# add one extra header for the POSCAR
headers.append("POSCAR_url")

# grab all the data from the rest of the rows
data = []
for row in tqdm(rows[1:]):
    row_data = []

    for column in row.find_elements_by_tag_name("td"):
        row_data.append(column.text)

    # we need to grab the POSCAR url too
    # which we can get indirectly from the link within the first column
    href = row.find_element_by_tag_name("a").get_attribute("href")
    poscar_url = href.replace(".html", ".poscar").replace(
        "/CrystalDatabase/", "/CrystalDatabase/POSCAR/"
    )
    # append to the row data and we will go through these later - rather switch back and forth from this main page
    row_data.append(poscar_url)

    data.append(row_data)


# add a POSCAR header
headers.append("POSCAR")
# add POSCARs to each data row
for entry in tqdm(data):

    try:  # some POSCARs don't have a page... not sure why. I should contact their devs. see A_oC8_64_f.I (#179)

        # grab prototype POSCAR url
        poscar_url = entry[9]

        # move to the webpage for its POSCAR
        driver.get(poscar_url)

        # download the POSCAR string (note '\n' breaks will be there, but that's needed)
        textElement = driver.find_element_by_tag_name("pre")
        poscar = textElement.text

        # append to results
        entry.append(poscar)
    except:
        entry.append(None)


# make DataFrame
df = pd.DataFrame(data=data, columns=headers)

# close the window
driver.close()

# export to csv file
df.to_csv("aflow_prototypes.csv")

##############################################################################

# # if you want to go from a str in POSCAR format to a pymatgen structure
# # Easier method would use Structure.from_file() but we currently have just a string
# from pymatgen.io.vasp import Poscar
# input_string = textElement.text
# Poscar.from_string(input_string, False,
#                                    read_velocities=False).structure

##############################################################################



# -*- coding: utf-8 -*-

from simmate.website.core_components.filters import (
    Structure,
    Forces,
    Thermodynamics,
    Calculation,
)
from simmate.database.base_data_types.relaxation import (
    Relaxation as RelaxationTable,
    IonicStep as IonicStepTable,
)


class Relaxation(Structure, Calculation):
    class Meta:
        model = RelaxationTable
        fields = dict(
            volume_change=["range"],
            band_gap=["exact", "range"],
            is_gap_direct=["exact"],
            energy_fermi=["range"],
            conduction_band_minimum=["range"],
            valence_band_maximum=["range"],
            **Structure.get_fields(),
            **Calculation.get_fields(),
        )


class IonicStep(Structure, Forces, Thermodynamics):
    class Meta:
        model = IonicStepTable
        fields = dict(
            **Structure.get_fields(),
            **Thermodynamics.get_fields(),
            **Forces.get_fields(),
        )



# -*- coding: utf-8 -*-

from simmate.toolkit.transformations.base import Transformation


class CoordinateOrderedPerturbation(Transformation):

    # known as "coordinate mutation" in USPEX
    # site locations are mutated where sites with lower order have higher preference for mutation
    # https://uspex-team.org/static/file/USPEX-LargeComplexSystems-2010.pdf
    # ""Coordinate mutation was found [2] to be ineffective, because “blind” displacement of the
    # atoms is much more likely to decrease the quality of a structure than to increase it.""
    #!!! because of the quote above, the coordinate displacement is order-dependent

    io_scale = "one_to_one"
    ninput = 1
    use_multiprocessing = False

    def __init__(
        self,
    ):
        pass

    def apply_transformation(self, structure, max_attempts=100):
        return



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.base import VaspTask
from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
)


class Quality02Relaxation(VaspTask):
    """
    Runs a rough VASP geometry optimization.

    `Quality 02` indicates that on a scale from 00 to 04, these are ranked 02 in
    quality (with 04 being the highest quality).

    Typically, you'd only want to run this relaxation on structures that were
    randomly created (and thus likely very unreasonable). More precise relaxations
    should be done afterwards. Therefore, instead of using this calculation,
    we recommend only using the relaxation/staged workflow, which uses this
    calculation as a third step.
    """

    # This uses the PBE functional with POTCARs that have lower electron counts
    # and convergence criteria when available.
    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS_LOW_QUALITY

    # because this calculation is such a low quality we don't raise an error
    # if the calculation fails to converge
    confirm_convergence = False

    # Make the unitcell relatively cubic before relaxing
    pre_sanitize_structure = True

    # These are all input settings for this task.
    incar = dict(
        # These settings are the same for all structures regardless of composition.
        PREC="Normal",
        EDIFF=1e-3,
        EDIFFG=1e-2,
        ENCUT=375,  # !!! Should this be based on the element type?
        ISIF=3,
        NSW=100,
        IBRION=2,  # for cases of bad starting sites
        POTIM=0.02,
        LCHARG=False,
        LWAVE=False,
        KSPACING=0.5,
        # The type of smearing we use depends on if we have a metal, semiconductor,
        # or insulator. So we need to decide this using a keyword modifier.
        multiple_keywords__smart_ismear={
            "metal": dict(
                ISMEAR=1,
                SIGMA=0.06,
            ),
            "non-metal": dict(
                ISMEAR=0,
                SIGMA=0.05,
            ),
        },
    )



# -*- coding: utf-8 -*-

import os
import yaml

from prefect.states import Completed

from simmate.calculators.vasp.inputs import Potcar
from simmate.workflow_engine import Workflow

from simmate.command_line.workflows import workflows
from simmate.conftest import make_dummy_files


def test_workflows_list_all(command_line_runner):
    # list the workflows
    result = command_line_runner.invoke(workflows, ["list-all"])
    assert result.exit_code == 0


def test_workflows_show_config(command_line_runner):

    # list the config for one workflow
    result = command_line_runner.invoke(
        workflows,
        ["show-config", "static-energy.vasp.mit"],
    )
    assert result.exit_code == 0

    # ensure failure when a nested workflow is given
    result = command_line_runner.invoke(
        workflows,
        ["show-config", "relaxation.vasp.staged"],
    )
    assert result.exit_code == 1

    # ensure the flow fails when a incorrect name is given
    result = command_line_runner.invoke(
        workflows,
        ["show-config", "non-existant-flow"],
    )
    assert result.exit_code == 1


def test_workflows_explore(command_line_runner):

    # Make sure it passes when args match
    result = command_line_runner.invoke(
        workflows,
        ["explore"],
        input="6\n2\n",  # gives 6 and then 2 for prompts
    )
    assert result.exit_code == 0

    # Make sure it fails when args don't match
    result = command_line_runner.invoke(
        workflows,
        ["explore"],
        input="1\n99\n",  # gives 1 and then 99 for prompts
    )
    assert result.exit_code == 1


def test_workflows_setup_only(command_line_runner, structure, mocker, tmpdir):

    # establish filenames
    cif_filename = os.path.join(tmpdir, "test.cif")
    new_dirname = os.path.join(tmpdir, "inputs")

    # TODO: switch out the tested workflow for one that doesn't require
    # VASP. As-is, I need to pretend to add a POTCAR file
    potcar_filename = os.path.join(new_dirname, "POTCAR")
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # write the structure to file to be used
    structure.to("cif", cif_filename)

    # now try writing input files to the tmpdir
    result = command_line_runner.invoke(
        workflows,
        [
            "setup-only",
            "static-energy.vasp.mit",
            "--structure",
            cif_filename,
            "--directory",
            new_dirname,
        ],
    )
    assert result.exit_code == 0
    assert os.path.exists(new_dirname)

    # ensure failure when a nested workflow is given
    result = command_line_runner.invoke(
        workflows,
        ["setup-only", "relaxation.vasp.staged", cif_filename],
    )
    assert result.exit_code == 1


def test_workflows_run(command_line_runner, structure, mocker, tmpdir):

    # establish filenames
    cif_filename = os.path.join(tmpdir, "test.cif")
    new_dirname = os.path.join(tmpdir, "inputs")

    # write the structure to file to be used
    structure.to("cif", cif_filename)

    # I don't want to actually run the workflow, so I override the run method
    mocker.patch.object(
        Workflow,
        "run",
        return_value=Completed(),
    )

    # now try writing input files to the tmpdir
    result = command_line_runner.invoke(
        workflows,
        [
            "run",
            "static-energy.vasp.mit",
            "--structure",
            cif_filename,
            "--directory",
            new_dirname,
        ],
    )
    assert result.exit_code == 0
    Workflow.run.assert_called_with(
        structure=cif_filename,
        directory=new_dirname,
    )

    # ensure failure on improperly matched kwargs
    result = command_line_runner.invoke(
        workflows,
        ["run", "static-energy.vasp.mit", "hangingkwarg"],
    )
    assert result.exit_code == 1


def test_workflows_run_cloud(command_line_runner, structure, mocker, tmpdir):

    # establish filenames
    cif_filename = os.path.join(tmpdir, "test.cif")
    new_dirname = os.path.join(tmpdir, "inputs")

    # write the structure to file to be used
    structure.to("cif", cif_filename)

    # I don't want to actually run the workflow, so I override the run method
    mocker.patch.object(
        Workflow,
        "run_cloud",
        return_value=Completed(),
    )

    # now try writing input files to the tmpdir
    result = command_line_runner.invoke(
        workflows,
        [
            "run-cloud",
            "static-energy.vasp.mit",
            "--structure",
            cif_filename,
            "--directory",
            new_dirname,
        ],
    )
    assert result.exit_code == 0
    Workflow.run_cloud.assert_called_with(
        structure=cif_filename,
        directory=new_dirname,
    )


def test_workflows_run_yaml(command_line_runner, structure, mocker, tmpdir):

    # establish filenames
    cif_filename = os.path.join(tmpdir, "test.cif")
    yaml_filename = os.path.join(tmpdir, "test.yaml")
    new_dirname = os.path.join(tmpdir, "inputs")

    # write the structure to file to be used
    structure.to("cif", cif_filename)

    # write the yaml file with our input args
    input_args = dict(
        structure=cif_filename,
        workflow_name="static-energy.vasp.mit",
        directory=new_dirname,
    )
    with open(yaml_filename, "w") as file:
        yaml.dump(input_args, file)

    # I don't want to actually run the workflow, so I override the run method
    mocker.patch.object(
        Workflow,
        "run",
        return_value=Completed(),
    )

    # now try writing input files to the tmpdir
    result = command_line_runner.invoke(
        workflows,
        ["run-yaml", yaml_filename],
    )
    assert result.exit_code == 0
    Workflow.run.assert_called_with(
        structure=cif_filename,
        directory=new_dirname,
    )

    # TODO: other yaml files to test with I would like to test these but the
    # current issues is that they are reliant on a vasp command. Maybe I need
    # to mock a lower level method like S3Task.run...?

    # A customized workflow
    """
    # Indicates we want to change the settings, using a specific workflow as a starting-point
    workflow_name: customized/vasp
    workflow_base: static-energy/mit
    
    # These would update the class attributes for the single workflow run
    # The "custom__" start indicates we are updating some attribute
    custom__incar: 
        ENCUT: 600
        KSPACING: 0.25
        MAGMOM: 0.9
    custom__potcar_mappings:
        Y: Y_sv
    
    # Then the remaining inputs are the same as the workflow_base
    structure: Y2CF2.cif
    command: mpirun -n 5 vasp_std > vasp.out
    """

    # From structure file
    """
    workflow_name: static-energy/mit
    structure: Y2CF2.cif
    command: mpirun -n 5 vasp_std > vasp.out
    """

    # From database structure
    """
    workflow_name: static-energy/mit
    structure:
        database_table: MITStaticEnergy
        database_id: 1
    command: mpirun -n 5 vasp_std > vasp.out
    """

    # Nested workflow
    """
    workflow_name: electronic-structure/matproj
    structure: Y2CF2.cif
    command: mpirun -n 5 vasp_std > vasp.out
    """



# -*- coding: utf-8 -*-

import os

import pytest

from simmate.conftest import copy_test_files, make_dummy_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import Frozen


def test_frozen(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="frozen",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")
    out_filename = os.path.join(tmpdir, "vasp.out")
    chgcar_filename = os.path.join(tmpdir, "CHGCAR")
    wavecar_filename = os.path.join(tmpdir, "WAVECAR")

    # We use a negative timeout to ensure this class fails
    error_handler = Frozen(timeout_limit=-1)

    # Confirm an error IS found
    assert error_handler.check(tmpdir) == True

    # Make first attempt at fixing the error (with IMIX present)
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmpdir)
    assert fix == "Removed IMIX=1 and deleted CHGCAR and WAVECAR"
    assert not Incar.from_file(incar_filename).get("IMIX", None)
    assert not os.path.exists(chgcar_filename)
    assert not os.path.exists(wavecar_filename)

    # Make second attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ALGO from Fast to Normal"
    assert Incar.from_file(incar_filename)["ALGO"] == "Normal"

    # Make third attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched SYMPREC from 1e-5 to 1e-8"
    assert Incar.from_file(incar_filename)["SYMPREC"] == 1e-08

    # Make final attempt at fixing the error, which raises an error
    with pytest.raises(Exception):
        fix = error_handler.correct(tmpdir)

    # Confirm an error IS NOT found when no file exists
    os.remove(out_filename)
    assert error_handler.check(tmpdir) == False



# -*- coding: utf-8 -*-

from .matproj_full import ElectronicStructure__Vasp__MatprojFull

# NOTE: we do not import the independent band structure and density of states
# workflows as these should not be called directly.



# -*- coding: utf-8 -*-

from django import forms
from django.core.exceptions import ValidationError

from simmate.toolkit import Structure
from simmate.workflow_engine import Workflow

# TODO:
# class SubmitWorkflowYaml(forms.ModelForm):
#     pass


class SubmitWorkflow(forms.Form):

    # All workflows, regardless of their other parameters, will always have a
    # "labels" input. This is a list of labels to submit the workflow with.
    labels = forms.CharField(initial="WarWulf")

    def clean(self):
        # This method cleans multiple fields that depend on eachother.
        # See https://docs.djangoproject.com/en/4.0/ref/forms/validation/
        cleaned_data = super().clean()

        # For each of the fields in the file-or-json mapping, the user must
        # supply one of the two and NOT both. This for loop goes these fields
        # that require this and ensure this has been completed.
        actual_fields = self.base_fields.keys()
        for possible_field in INPUT_MAPPINGS["file-or-json"]:
            if f"{possible_field}_file" in actual_fields:
                file_input = cleaned_data.pop(f"{possible_field}_file", None)
                json_input = cleaned_data.pop(f"{possible_field}_json", None)

                if not file_input and not json_input:
                    raise ValidationError(
                        f"An input for {possible_field} is required. Please either "
                        "provide a file or json input for this parameter."
                    )
                elif file_input and json_input:
                    raise ValidationError(
                        f"Only one input for {possible_field} is allowed. Please either "
                        "provide a file or json input for this parameter -- not both."
                    )

                # lastly we want to remove these two fields and set the final
                # one to the parameter name (e.g. "structure_file" will be
                # changed to kwarg "structure")
                cleaned_data[possible_field] = file_input if file_input else json_input

        return cleaned_data

    def clean_structure_json(self, field_name: str = "structure_json"):

        # Grab what the user submitted
        structure_json = self.cleaned_data[field_name]

        if not structure_json:
            # This means not structure_file was provided and the user should
            # have supplied a structure_json instead.
            return

        try:
            structure = Structure.from_dynamic(structure_json)
        except:
            raise ValidationError("Failed to load the JSON input provided.")

        return structure

    def clean_structure_file(self, field_name: str = "structure_file"):

        # Grab what the user submitted
        structure_file = self.cleaned_data[field_name]

        # Make sure the file isn't too large
        if structure_file:
            # 5e6 is 5MB limit
            if structure_file.size > 5e6:
                raise ValidationError("File upload is too large (>5MB)")
        else:
            # This means not structure_file was provided and the user should
            # have supplied a structure_json instead.
            return

        # If we make it through the checks above, we can now convert the
        # text of the file into a pymatgen structure object
        structure_str = structure_file.read().decode("utf-8")

        # OPTIMIZE: is there a cleaner way to determine format?
        formats = ["cif", "poscar", "cssr", "json", "yaml", "xsf", "mcsqs"]
        for fmt in formats:
            try:
                structure = Structure.from_str(structure_str, fmt=fmt)
            except:
                structure = None

            # break the for-loop once structure is succesfully loaded
            if structure:
                break

        # ensure we loaded the structure successfully
        if not structure:
            raise ValidationError(
                f"Fail to read the file provided. Supported formats are... {formats}"
            )

        return structure

    def clean_structure_start_json(self):
        return self.clean_structure_json("supercell_start_json")

    def clean_structure_end_json(self):
        return self.clean_structure_json("supercell_end_json")

    def clean_migration_images_json(self):
        return self.clean_structure_json("migration_images_json")

    def clean_structure_start_file(self):
        return self.clean_structure_file("supercell_start_file")

    def clean_structure_end_file(self):
        return self.clean_structure_file("supercell_end_file")

    def clean_migration_images_file(self):
        return self.clean_structure_file("migration_images_file")

    def clean_labels(self):

        # Grab what the user submitted
        labels_str = self.cleaned_data["labels"]

        # convert the string to a list of labels.
        # For example...
        #   "WarWulf, LongLeaf, DogWood"
        #       converts to
        #   ['WarWulf', ' LongLeaf', ' DogWood']
        labels = labels_str.strip().split(",")

        return labels

    @classmethod
    def from_workflow(cls, workflow: Workflow):

        # We keep a running dictionary of options for the form.
        form_fields = {}
        for parameter_name in workflow.parameter_names:

            is_required = parameter_name in workflow.parameter_names_required

            if parameter_name in INPUT_MAPPINGS["file-or-json"]:
                # Only one of these two inputs is required. I set required=False
                # for both, but I should check that one is given within this
                # class's "clean" method

                # Max length refers to the filename length, not its contents
                form_fields[f"{parameter_name}_file"] = forms.FileField(
                    max_length=100,
                    required=False,
                )
                form_fields[f"{parameter_name}_json"] = forms.JSONField(required=False)

            elif parameter_name in INPUT_MAPPINGS["string"]:
                form_fields[parameter_name] = forms.CharField(
                    max_length=100,
                    required=is_required,
                )

            elif parameter_name in INPUT_MAPPINGS["integer"]:
                form_fields[parameter_name] = forms.IntegerField(required=is_required)

            elif parameter_name in INPUT_MAPPINGS["float"]:
                form_fields[parameter_name] = forms.FloatField(required=is_required)

            elif parameter_name in INPUT_MAPPINGS["boolean"]:
                form_fields[parameter_name] = forms.BooleanField(required=is_required)

            elif parameter_name in INPUT_MAPPINGS["json"]:
                form_fields[parameter_name] = forms.JSONField(required=is_required)

            else:
                raise Exception(f"Unknown input type for parameter {parameter_name}")

        # once we have all the fields, we can dynamically create the new class.
        NewClass = type("SubmitWorkflowForm", tuple([cls]), form_fields)

        return NewClass


# TODO: is there a way to dynamically inpsect types so that I don't need this?
# I need a way to accept filters from custom workflows as well. How should
# I handle user-based parameters?
INPUT_MAPPINGS = {
    "json": [
        "source",
        "migration_hop",
    ],
    "file-or-json": [
        "structure",
        "supercell_start",
        "supercell_end",
        "migration_images",
    ],
    "string": [
        "directory",
        "directory_old",
        "directory_new",
        "command",
        "migrating_specie",
    ],
    "integer": [
        "nsteps",
        "diffusion_analysis_id",
        "migration_hop_id",
    ],
    "float": [
        "temperature_end",
        "temperature_start",
        "time_step",
    ],
    "boolean": [
        "copy_previous_directory",
        "pre_standardize_structure",
        "pre_sanitize_structure",
        "is_restart",
    ],
}



# -*- coding: utf-8 -*-

from django.urls import path

from simmate.website.workflows import views

urlpatterns = [
    #
    # Lists off the different types of calculations
    path(
        route="",
        view=views.workflows_all,
        name="workflows",
    ),
    #
    # Lists off the different workflows of a given type
    path(
        route="<workflow_type>/",
        view=views.workflows_by_type,
        name="workflows_by_type",
    ),
    #
    # A detailed view of a specific workflow. This page will contain sections
    # about the workflow as well as host actions for submitting and querying
    # new flow runs.
    path(
        route="<workflow_type>/<workflow_name>/",
        view=views.WorkflowAPIViewSet.dynamic_list_view,
        name="workflow_detail",
    ),
    #
    # Views results for an individual calculation
    path(
        route="<workflow_type>/<workflow_name>/<int:pk>",
        view=views.WorkflowAPIViewSet.dynamic_retrieve_view,
        name="workflow_run_detail",
    ),
    #
    # Submit a new calculation
    path(
        route="<workflow_type>/<workflow_name>/submit",
        view=views.workflow_submit,
        name="workflow_submit",
    ),
]



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import table_column, Calculation

from typing import List


class NestedCalculation(Calculation):
    """
    A nested calculation is extremely similar to a calculation table, except here
    the linked workflow is a NestedWorkflow -- meaning it is made of multiple
    smaller workflows. For example, we may want to run a workflow that runs a
    series of relaxations. Or maybe a relaxation, then energy, then band-strucuture
    calculation. This table is for keeping track of workflows ran in series like
    this.

    Typically, you'll use the `create_subclass_from_calcs` method to create a
    subclass of this table, which handles making all the relationships for you.
    """

    class Meta:
        abstract = True
        app_label = "workflows"

    # TODO:
    # child_database_tables = [...]

    # TODO:
    # should this be a list of all modifications? It could maybe be used to
    # carry fixes (such as smearing) accross different calcs.
    # Or maybe a method that just lists the corrections of each subcalc?
    corrections = None
    # For now I delete this column. This line removes the field.

    @classmethod
    def create_subclass_from_calcs(
        cls,
        name: str,
        child_database_tables: List[Calculation],
        module: str,
        **extra_columns,
    ):
        """
        Dynamically creates a subclass of NestedCalculation -- and handles linking
        together all child calculation tables.

        `simmate.calculators.vasp.database.relaxation` shows an example of creating
        a table from this class:

        ``` python
        StagedRelaxation = NestedCalculation.create_subclass_from_calcs(
            "StagedRelaxation",
            [
                Quality00Relaxation,
                Quality01Relaxation,
                Quality02Relaxation,
                Quality03Relaxation,
                Quality04Relaxation,
            ],
            module=__name__,
        )
        ```

        To add custom columns, you can do the following:

        ``` python
        from simmate.database.base_data_types import table_column

        StagedRelaxation = NestedCalculation.create_subclass_from_calcs(
            ... # everything is the same as above
            custom_column_01=table_column.FloatField()
        )
        ```

        #### Parameters

        - `name` :
            Name of the subclass that is output.
        - `child_database_tables` :
            list of database tables for the nested workflows. This table links
            these sub-tables together so results can be viewed from each step.
        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.
        **extra_columns : TYPE
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        #### Returns

        NewClass :
            A subclass of NestedCalculation.

        """

        # BUG: I assume a workflow won't point to the save calculation table
        # more than once... What's a scenario where this isn't true?
        # I can only think of multi-structure workflows (like EvolutionarySearch)
        # which I don't give their own table for now.
        new_columns = {}
        for child_calc in child_database_tables:
            new_column = table_column.OneToOneField(
                child_calc,
                on_delete=table_column.CASCADE,
                # related_name=...,
                blank=True,
                null=True,
            )
            new_columns[f"{child_calc._meta.model_name}"] = new_column

        # Now put all the fields together to make the new class
        NewClass = cls.create_subclass(
            name,
            **new_columns,
            **extra_columns,
            # also have child calcs list as an attribute
            child_database_tables=child_database_tables,
            module=module,
        )

        # we now have a new child class and avoided writing some boilerplate code!
        return NewClass

    # def update_calculation(self):
    #     """
    #     This is an experimental method that iterates through child workflow tables
    #     and links the results to the NestedCalculation
    #     """
    #     # BUG: This assumes we ran all calculations within the same directory,
    #     # which isn't true in all cases.
    #     for child_calc_table in self.child_database_tables:
    #         if child_calc_table.objects.filter(directory=self.directory).exists():
    #             child_calc = child_calc_table.objects.get(directory=self.directory)
    #             setattr(self, child_calc._meta.model_name, child_calc)
    #     self.save()



# -*- coding: utf-8 -*-

from simmate.toolkit import Structure

# TODO: Maybe make an USPEX calculator

#!!! NOT TESTED
class USPEXStructure:

    # This wrapper tricks USPEX into thinking its submitting structures, when really its just making structures then quitting.
    #!!! In the future, USPEX devs will hopefully have an accessible command to just create structures. This will dramatically improve
    #!!! both the speed of this wrapper as well as the cleaniness of the code.
    # Because this wrapper needs to trick USPEX, it is very slow at calling new_structure. It's much more efficient to trick USPEX a single
    # time and thus make a bunch of structures at once (new_structures). This is because of the overhead of calling USPEX and having it generate
    # input files for a calculation which is never going to actually run. Once USPEX has a convience function for making structures, this can be undone.
    #!!! Also consider reading the .mat files directly so I don't have to waste time generating input folders.
    # https://scipy-cookbook.readthedocs.io/items/Reading_mat_files.html

    # see source: https://uspex-team.org/en/uspex/downloads
    # see tutorials: NONE

    def __init__(
        self,
        composition,  # must be a pymatgen Composition object
        uspex_python_env="uspex",  # This needs to be a custom python env! It must be Python=2.7 and have Numpy, Scipy, MatPlotLib, ASE, and SpgLib
        uspex_loc="/home/jacksund/USPEX",  # Location where USPEX was installed to -- the command itself doesn't work... I need to figure out what's going on
        conda_loc="/home/jacksund/anaconda3/etc/profile.d/conda.sh",  # this will vary if you didn't install anaconda to your home directory or installed miniconda instead.
        temp_dir="/home/jacksund/Desktop/uspex_tmp",  # this is the temporary directory where I will run uspex
    ):

        self.uspex_python_env = uspex_python_env
        self.temp_dir = temp_dir

        #!!! this is inside the init because not all users will have this installed!
        # see if the user has USPEX installed
        import subprocess

        #!!! what is the best way to see if AIRSS is installed? Check the path?
        output = subprocess.run(
            "USPEX --help",  # command that calls USPEX
            shell=True,  # use commands instead of local files
            capture_output=True,  # capture command output
            text=True,  # convert output from bytes to string
        )
        #!!! THIS DOESN'T WORK AT THE MOMENT...
        # if output.returncode != 0:
        #     #!!! I should raise an error in the future
        #     print('You must have USPEX installed to use USPEXStructure!!')
        #     return # exit the function as the script will fail later on otherwise

        #!!! add check that the user is on Linux?

        #!!! add check that the user has a proper python enviornment set up in conda?

        # Format composition in the manner USPEX requests it.
        # This is a list of atomic symvols (for example, MgSiO3 is 'Mg Si O')
        atom_types = " ".join([element.symbol for element in composition])
        # This is a list of non-reduced species count.
        # If I want variable counts, I would use minAt and maxAt settings + the reduced formula, but I don't do that
        num_species = " ".join(
            [str(int(composition[element])) for element in composition]
        )

        # Now let's set all of the other USPEX options
        #!!! consider moving default inputs to the init keyword options
        uspex_options = {
            "calculationType": "300",  # see USPEX docs for more option - for example, set to 200 for 2-D structures
            "symmetries": "2-230",  #!!! how do I make this compatible with spacegroup_options...?
            "fracTopRand": "0.0",  # fraction Topological structures to make #!!! I should keep this as 0 or 1 for clarity within the PyMatDisc code
            "atom_type": atom_types,
            "num_species": num_species,
        }

        # Here we setup the input file to create USPEX structures.
        # Whereever you see some name between brackets {var}, we will replace with the dict above.
        # NUM_STRUCTURES we replace in functions below.
        # Note that the commandExecutable is just a dummy echo command and doesn't do anything.
        # We also set the calc to VASP because we want POSCAR files made for us.
        uspex_input = """
{calculationType} : calculationType

% symmetries
{symmetries}
% endSymmetries

% atomType
{atom_type}
% EndAtomType

% numSpecies
{num_species}
% EndNumSpecies

{fracTopRand} : fracTopRand

NUM_STRUCTURES : initialPopSize

% abinitioCode
1
% ENDabinit

% commandExecutable
echo skip this
% EndExecutable

NUM_STRUCTURES : numParallelCalcs

1 : whichCluster
"""

        # Now format this input string with our options. The only remaining input to set is the number
        # of structures to generate (NUM_STRUCTURES)
        self.uspex_input = uspex_input.format(**uspex_options)

        # In order to have USPEX create structures, it is also going to make input files for VASP.
        # To do this, there needs to be a Specific folder with INCAR_1 and POTCAR_X (X=symbol)
        # Because we aren't really running VASP, these don't need to be legit files. In fact, making them
        # empty is easier on the computer because we don't have to repeated paste large POTCAR files.
        # Let's make those dummy files here.

        # First let's switch to the temp_dir and save the current working dir (cwd) for reference
        import os

        cwd = os.getcwd()
        os.chdir(temp_dir)  #!!! What if dir doesn't exist yet? This will throw an error

        # make the Specifics folder and move into it
        os.mkdir("Specific")
        os.chdir("Specific")

        # Make the INCAR_1 file.
        subprocess.run(
            "echo DUMMY INCAR > INCAR_1",
            shell=True,  # use commands instead of local files
        )

        # Make the POTCAR_X files.
        for element in composition:
            subprocess.run(
                "echo DUMMY POTCAR > POTCAR_{}".format(element.symbol),
                shell=True,  # use commands instead of local files
            )

        # move back to the temp dir
        os.chdir("..")

        # make the Submission folder and move into it
        os.mkdir("Submission")
        os.chdir("Submission")

        # make the submitJob_local.py file
        localsubmit_content = """
from __future__ import with_statement
from __future__ import absolute_import
from subprocess import check_output
import re
import sys
from io import open

def submitJob_local(index, commnadExecutable):
    return 12345 #

if __name__ == u'__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(u'-i', dest=u'index', type=int)
    parser.add_argument(u'-c', dest=u'commnadExecutable', type=unicode)
    args = parser.parse_args()

    jobNumber = submitJob_local(index=args.index, commnadExecutable=args.commnadExecutable)
    print('<CALLRESULT>')
    print(int(jobNumber))
"""
        # write it and close immediately
        file = open("submitJob_local.py", "w")
        file.writelines(localsubmit_content)
        file.close()
        # go back up a directory
        os.chdir("..")

        # The last issue is that USPEX needs a different python enviornment, which we need to access via a script.
        # Let's add a script run_uspex.sh that sets this env for us and then runs uspex.
        submit_content = """
source {conda_loc}
conda activate {uspex_python_env}
export USPEXPATH={uspex_loc}/application/archive/src
export MCRROOT={uspex_loc}
{uspex_loc}/application/archive/USPEX -r
"""
        # write it and close immediately
        file = open("run_uspex.sh", "w")
        file.writelines(
            submit_content.format(
                **{
                    "conda_loc": conda_loc,
                    "uspex_python_env": uspex_python_env,
                    "uspex_loc": uspex_loc,
                }
            )
        )
        file.close()
        # give permissions to the script so we can run it below
        subprocess.run(
            "chmod a+x run_uspex.sh",  #!!! is a+x correct?
            shell=True,  # use commands instead of local files
        )

        # We now have everything except the INPUT.txt file! We will write that and run it below.
        # This is done later because we don't know NUM_STRUCTURES yet

        # switch back to our original working dir
        os.chdir(cwd)

    def new_structures(self, n):

        # See my comments above on why this atypical function exists... (it's much faster than calling USPEX each new struct)

        # First let's switch to the temp_dir and save the current working dir (cwd) for reference
        import os

        cwd = os.getcwd()
        os.chdir(self.temp_dir)

        # make the INPUT.txt file with n as our NUM_STRUCTURES to make
        # write it and close immediately
        file = open("INPUT.txt", "w")
        file.writelines(self.uspex_input.replace("NUM_STRUCTURES", str(n)))
        file.close()

        # now let's have USPEX run and make the structures!
        import subprocess

        output = subprocess.run(
            "bash run_uspex.sh",  #!!! should I pipe the output to a log file?
            shell=True,  # use commands instead of local files
            #!!! comment out the options below if you want USPEX to print to terminal
            capture_output=True,  # capture command output
            text=True,  # convert output from bytes to string
        )

        # All the structures are put into folders as POSCAR files. The folders are CalcFold1, CalcFold2, ... CalcFold200...
        # Let's iterate through these and pool them into a list
        import shutil  # to remove directories #!!! is there another way to do this?

        structures = []
        for i in range(
            n
        ):  # we can assume all folders are there instead of grabbing os.listdir() for all CalcFold*
            os.chdir(
                "CalcFold{}".format(i + 1)
            )  # the +1 is because we don't want to count from 0
            structure = Structure.from_file(
                "POSCAR"
            )  # convert POSCAR to pymatgen Structure
            structures.append(structure)  # add it to the list
            os.chdir("..")  # move back directory
            shutil.rmtree(
                "CalcFold{}".format(i + 1)
            )  # delete the folder now that we are done with it

        # Do some cleanup and delete all the unneccesary directories that were just made.
        # This sets us up to run new_structures again
        # let's go through the folders first
        rm_dir_list = ["AntiSeeds", "results1", "Seeds", "CalcFoldTemp"]
        for d in rm_dir_list:
            shutil.rmtree(d)
        # Rather than go through a list like with the directories, it's easier
        # to just delete all .mat files because that's what they all are
        subprocess.run(
            "rm *.mat*",
            shell=True,  # use commands instead of local files
        )

        # switch back to our original working dir
        os.chdir(cwd)

        # return the list of pymatgen Structure objects that we've made
        return structures

    def new_structure(self, spacegroup=None, lattice=None, species=None, coords=None):

        # IF YOU WANT TO MAKE MANY STRUCTURES, THIS IS VERY SLOW! Use new_structures(n) instead.

        # if the user supplied a sites, it's going to be overwritten.
        if coords or species:
            #!!! this error will always show up in my current form of sample.py -- should I change that?
            #!!! the only time the error doesn't show up is when no validators fail on the object
            print(
                "Warning: While USPEX allows for specifying a fixed lattice, "
                "our wrapper does not currently support this. Contact our team if you"
                " would like this feature incorporated."
            )

        # if the user supplied a lattice, it's going to be overwritten.
        if lattice:
            print(
                "Warning: While USPEX allows for specifying a fixed lattice, "
                "our wrapper does not currently support this. Contact our team if you"
                " would like this feature incorporated."
            )

        # if a spacegroup is not specified, grab a random one from our options
        # no check is done to see if the spacegroup specified is compatible with the vector_generator built
        if spacegroup:
            print(
                "Warning: While USPEX allows for specifying a spacegroup(s), "
                "our wrapper does not currently support this. Contact our team if you"
                " would like this feature incorporated."
            )

        # call the new_structures() function and tell it to create just one structure
        structure = self.new_structures(1)[0]

        return structure



# -*- coding: utf-8 -*-

from simmate.materials_science.elements import Element


class Species:
    """

    Species is an central class to using Simmate as there are many types that we
    need to account for. We use https://en.wikipedia.org/wiki/Chemical_species as
    a template for the type of species:
        atomic/elemental species --> This is just an Element with extra data
        ionic species --> an elemental species with an oxidation state
        molecular species -->  a moleucle object with extra data
        radical species --> species with unpaired electron
        compositional species --> similar to elemental species but with partial occupancy
            or with multiple elements & stoichiometric ratios



    Note all species can have extra data attached to it, but this *DOES NOT* include
    coordinates. This is generic data such as an ideal oxidation state or spin data.

    This class can be confusing when compared to Sites, which are Elements with
    extra data attached to each and *DOES* include coordinates.

    Deciding where to attach data (either Species or Site) can be tricky, but as
    a general rule, you should always use Site when working with Molecules and
    Structures -- and then use Species when you're working generically with a
    Composition.

    As an example, say I want to predict oxidation states. For the material Ca2NCl,
    I would make a Composition object made of Species, and each Species would have
    an oxidation state attached to it. For a crystal structure of Ca2NCl, I would
    want to instead attach oxidation states to each Site.

    As a bonus, if I wanted to switch from a Structure (made of Sites) to a generic
    Composition (made of Species), I can use a Structure's composition attribute to
    get a Composition object. This would group all of the Sites by their unique
    properties (ignoring coordinates) and give you a list of the unique Species.

    For extra help, use this tutorial or you can ask our team about your situation:
        < TODO: ADD LINK >
    """

    def __init__(self, symbol, oxidation_state=0.0, **kwargs):

        # Because oxidation states are the most commonly used with Species objects,
        # I set that keyword, but all others can be defined by the user. Here are
        # some standard keywords worth considering:
        #   oxidation_state
        #   spin_up / spin_down / spin_total
        #   population

        # for each additional keyword and value provided, link it to an attribute
        pass



# -*- coding: utf-8 -*-

from simmate.toolkit import Structure

# TODO: Maybe make a CALYPSO calculator

#!!! NOT TESTED
class CALYPSOStructure:

    # This wrapper runs CALYPSO in 'split mode', where structures are simply created and then no further steps are taken.
    #!!! In the future, CALYPSO devs will hopefully have an accessible command to just create structures. This will dramatically improve
    #!!! both the speed of this wrapper as well as the cleaniness of the code.

    # see source: http://www.calypso.cn/getting-calypso/
    # see tutorials: NONE

    def __init__(
        self,
        composition,  # must be a pymatgen Composition object
        calypso_exe_loc="/home/jacksund/Desktop/",  # location of the calypso.x file
        calypso_python_env="uspex",  #!!! CHANGE NAME # This needs to be a custom python env! It must be Python=2.7 and have Numpy, Scipy, MatPlotLib, ASE, and SpgLib
        conda_loc="/home/jacksund/anaconda3/etc/profile.d/conda.sh",  # this will vary if you didn't install anaconda to your home directory or installed miniconda instead.
        temp_dir="/home/jacksund/Desktop/calypso_tmp",  # this is the temporary directory where I will run uspex
    ):
        self.calypso_python_env = calypso_python_env
        self.temp_dir = temp_dir

        #!!! add check that the user is on Linux?

        #!!! add check that the user has a proper python enviornment set up in conda?

        #!!! Because of bug with calypso's poscar files, I need to save this
        self.comp = " ".join([element.symbol for element in composition])

        # Now let's set all of the USPEX options and format them the way they'd like
        #!!! consider moving default inputs to the init keyword options
        # There are many  more options (such as for 2D materials) that can be added
        calypso_options = {
            "NumberOfSpecies": len(composition),  # number of unique elements
            "NameOfAtoms": " ".join(
                [element.symbol for element in composition]
            ),  # This is a list of atomic symvols (for example, MgSiO3 is 'Mg Si O')
            "NumberOfAtoms": " ".join(
                [
                    str(int(composition[element]))
                    for element in composition.reduced_composition
                ]
            ),  # This is a list of reduced species count
            "NumberOfFormula": int(
                composition.num_atoms / composition.reduced_composition.num_atoms
            ),  # factor of formula unit -- for example Mg3Si3O9 would have 3 (vs MgSiO3)
        }

        #!!! CALYPSO's auto volume prediction doesn't look to work very well.
        #!!! It's either that, or the distance checks are too strict.
        # Therefore, let's predict volume on our own and input it.
        volume = composition.volume_estimate(packing_factor=2)
        calypso_options.update({"Volume": volume})

        # CALYPSO says they have a default value for distances, but it doesn't
        # look like they work. So I am going to end up making a distance
        # matrix here.
        dm = composition.distance_matrix_estimate()
        dm_str = ""
        for row in dm:
            for val in row:
                dm_str += str(val) + " "
            dm_str += "\n"
        dm_str = dm_str[:-2]  # remove the final \n
        calypso_options.update({"DistanceOfIon": dm_str})

        # Here we setup the input file to create CALYPSO structures.
        # Whereever you see some name between brackets {var}, we will replace with the dict above.
        # NUM_STRUCTURES we replace in functions below.
        # Note NumberOfFormula is input twice becuase we want a fixed min/max #!!! change this in the future...?
        calypso_input = """
NumberOfSpecies = {NumberOfSpecies}
NameOfAtoms = {NameOfAtoms}
NumberOfAtoms = {NumberOfAtoms}
NumberOfFormula = {NumberOfFormula} {NumberOfFormula}
PopSize = NUM_STRUCTURES
Command = echo SKIP THIS
Split = T
Volume = {Volume}
@DistanceOfIon
{DistanceOfIon}
@End
"""

        # Now format this input string with our options. The only remaining input to set is the number
        # of structures to generate (NUM_STRUCTURES)
        self.calypso_input = calypso_input.format(**calypso_options)

        # In order to have CALYPSO create structures, we need the calypso.x file copied into our directory

        # First let's switch to the temp_dir and save the current working dir (cwd) for reference
        import os

        cwd = os.getcwd()
        os.chdir(temp_dir)  #!!! What if dir doesn't exist yet? This will throw an error

        # Copy the calypso.x file into this directory
        import subprocess

        subprocess.run(
            "cp {}/calypso.x {}/".format(calypso_exe_loc, temp_dir),
            shell=True,  # use commands instead of local files
        )

        # The last issue is that CALYPSO needs a different python enviornment, which we need to access via a script.
        # Let's add a script run_uspex.sh that sets this env for us and then runs uspex.
        submit_content = """
source {conda_loc}
conda activate {calypso_python_env}
./calypso.x
"""
        # write it and close immediately
        file = open("run_calypso.sh", "w")
        file.writelines(
            submit_content.format(
                **{"conda_loc": conda_loc, "calypso_python_env": calypso_python_env}
            )
        )
        file.close()
        # give permissions to the script so we can run it below
        subprocess.run(
            "chmod a+x run_calypso.sh",  #!!! is a+x correct?
            shell=True,  # use commands instead of local files
        )

        # We now have everything except the INPUT.txt file! We will write that and run it below.
        # This is done later because we don't know NUM_STRUCTURES yet

        # switch back to our original working dir
        os.chdir(cwd)

    def new_structures(self, n):

        # See my comments above on why this atypical function exists... (it's much faster than calling USPEX each new struct)

        # First let's switch to the temp_dir and save the current working dir (cwd) for reference
        import os

        cwd = os.getcwd()
        os.chdir(self.temp_dir)

        # make the INPUT.txt file with n as our NUM_STRUCTURES to make
        # write it and close immediately
        file = open("input.dat", "w")
        file.writelines(self.calypso_input.replace("NUM_STRUCTURES", str(n)))
        file.close()

        # now let's have USPEX run and make the structures!
        import subprocess

        subprocess.run(
            "bash run_calypso.sh",  #!!! should I pipe the output to a log file?
            shell=True,  # use commands instead of local files
            #!!! comment out the options below if you want USPEX to print to terminal
            capture_output=True,  # capture command output
            text=True,  # convert output from bytes to string
        )

        # All the structures are as POSCAR files. The files are POSCAR_1, POSCAR_2, ... POSCAR_n...
        # Let's iterate through these and pool them into a list
        structures = []
        for i in range(
            n
        ):  # we can assume all folders are there instead of grabbing os.listdir() for all CalcFold*
            poscar_name = "POSCAR_{}".format(i + 1)

            #!!! BUG WITH CALYPSO...
            # They don't add the atom types to the POSCAR... wtf
            # I need to do that manually here
            file = open(poscar_name, "r")
            lines = file.readlines()
            file.close()
            file = open(poscar_name, "w")
            lines.insert(5, self.comp + "\n")  # self.comp
            file.writelines(lines)
            file.close()

            # now we can load the POSCAR and add it to our list
            structure = Structure.from_file(
                poscar_name
            )  # convert POSCAR to pymatgen Structure # the +1 is because we don't want to count from 0
            structures.append(structure)  # add it to the list
            os.remove(poscar_name)  # delete the file now that we are done with it

        # Do some cleanup and delete all the unneccesary directories/files that were just made.
        # This sets us up to run new_structures again
        import shutil  # to remove directories #!!! is there another way to do this?

        shutil.rmtree("results")
        # Rather than go through a list like with the directories, it's easier
        # to just delete all .mat files because that's what they all are
        subprocess.run(
            "rm *.py*",
            shell=True,  # use commands instead of local files
        )
        # There's also two more files to remove - POSCAR and step
        subprocess.run(
            "rm POSCAR; rm step",
            shell=True,  # use commands instead of local files
        )

        # switch back to our original working dir
        os.chdir(cwd)

        # return the list of pymatgen Structure objects that we've made
        return structures

    def new_structure(self, spacegroup=None, lattice=None, species=None, coords=None):

        # IF YOU WANT TO MAKE MANY STRUCTURES, THIS IS VERY SLOW! Use new_structures(n) instead.

        # if the user supplied a sites, it's going to be overwritten.
        if coords or species:
            #!!! this error will always show up in my current form of sample.py -- should I change that?
            #!!! the only time the error doesn't show up is when no validators fail on the object
            print(
                "Warning: While CALYPSO allows for specifying a fixed lattice, "
                "our wrapper does not currently support this. Contact our team if you"
                " would like this feature incorporated."
            )

        # if the user supplied a lattice, it's going to be overwritten.
        if lattice:
            print(
                "Warning: While CALYPSO allows for specifying a fixed lattice, "
                "our wrapper does not currently support this. Contact our team if you"
                " would like this feature incorporated."
            )

        # if a spacegroup is not specified, grab a random one from our options
        # no check is done to see if the spacegroup specified is compatible with the vector_generator built
        if spacegroup:
            print(
                "Warning: While CALYPSO allows for specifying a spacegroup(s), "
                "our wrapper does not currently support this. Contact our team if you"
                " would like this feature incorporated."
            )

        # call the new_structures() function and tell it to create just one structure
        structure = self.new_structures(1)[0]

        return structure



# -*- coding: utf-8 -*-

import os

from simmate.conftest import copy_test_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import TetrahedronMesh


def test_tetrahedron_mesh(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="tetrahedron_mesh",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")

    # init class with default settings
    error_handler = TetrahedronMesh()

    # Confirm an error IS found using test files
    assert error_handler.check(tmpdir) == True

    # Make first attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched KSPACING from 0.6 to 0.48"
    assert Incar.from_file(incar_filename)["KSPACING"] == 0.48

    # Make second attempt at fixing the error, for when we have the default
    # KSPACING value.
    incar = Incar.from_file(incar_filename)
    incar["KSPACING"] = 0.5
    incar.to_file(incar_filename)
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ISMEAR to 0 and SIGMA to 0.05"
    incar = Incar.from_file(incar_filename)
    assert incar["ISMEAR"] == 0
    assert incar["SIGMA"] == 0.05

    # Confirm an error IS NOT found when there is no vasp.out
    error_handler.filename_to_check = "vasp.does_not_exist"
    assert error_handler.check(tmpdir) == False



# -*- coding: utf-8 -*-

import os

from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.io.vasp.inputs import Kpoints

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class MeshSymmetry(ErrorHandler):
    """
    Corrects the mesh symmetry error in VASP. This error is sometimes
    non-fatal and the job can complete successfully. So this error handler
    only checks at the end of the run, and if the run has converged, no error
    is recorded. We can also ignore this error if symmetry is turned off or if
    automatic k-mesh has been used.
    """

    is_monitor = False
    filename_to_check = "vasp.out"
    possible_error_messages = [
        "Reciprocal lattice and k-lattice belong to different class of"
    ]

    def check(self, directory: str) -> bool:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # check if there is a KPOINTS file and if so, read it and check the
        # kpoint style being using.
        kpoints_filename = os.path.join(directory, "KPOINTS")
        if os.path.exists(kpoints_filename):
            kpoints = Kpoints.from_file(kpoints_filename)
            kpoints_style = kpoints.style
        else:
            kpoints_style = None

        # We can say there is no error if one of the following is true:
        #   (1) symmetry is turned off
        #   (2) KSPACING is used
        #   (3) KPOINTS file uses and "Automatic" mesh
        if (
            incar.get("ISYM", 1) == 0
            or incar.get("KSPACING")
            or kpoints_style == Kpoints.supported_modes.Automatic
        ):
            return False

        # Now check if the calculation converged. If it did, we ignore the error.
        xml_filename = os.path.join(directory, "vasprun.xml")
        try:
            # load the xml file and only parse the bare minimum
            xmlReader = Vasprun(
                filename=xml_filename,
                parse_dos=False,
                parse_eigen=False,
                parse_projected_eigen=False,
                parse_potcar_file=False,
                exception_on_bad_xml=True,
            )
            if xmlReader.converged:
                return False
        except Exception:
            pass

        # Finally, we can now search for the error message. We can use the parent
        # class's default function to do this.
        return super().check(directory)

    def correct(self, directory: str) -> str:

        raise NotImplementedError(
            "The fix for MeshSymmetryError hasn't been converted from Custodian "
            "to Simmate yet. A fix does exist though, so be sure to tell our "
            "that team you need it!"
        )

        # load the INCAR file to view the current settings
        kpoints_filename = os.path.join(directory, "KPOINTS")
        kpoints = Kpoints.from_file(kpoints_filename)

        # !!! I'm not sure what Custodian is doing here exactly, so I need to
        # revisit this when I have my KPOINTS class implemented

        # from functools import reduce
        # import operator
        # backup(VASP_BACKUP_FILES | {self.output_filename})
        # vi = VaspInput.from_directory(".")
        # m = reduce(operator.mul, vi["KPOINTS"].kpts[0])
        # m = max(int(round(m ** (1 / 3))), 1)
        # if vi["KPOINTS"].style.name.lower().startswith("m"):
        #     m += m % 2
        # actions = [{"dict": "KPOINTS", "action": {"_set": {"kpoints": [[m] * 3]}}}]
        # VaspModder(vi=vi).apply_actions(actions)
        # return {"errors": ["mesh_symmetry"], "actions": actions}



# -*- coding: utf-8 -*-

"""
This file is only required for editable installs using pip. We can
remove this file once the following issue is closed:
  https://github.com/pypa/setuptools/issues/2816

As extra, the example_project/setup.py can also be converted to pyproject.toml
and removed. I hold off on this to minimize new files made for users.
"""

from setuptools import setup

setup()



# -*- coding: utf-8 -*-

"""
This settings file is only for running Simmate's test suite with pytest.

It contain's a little "hack" so devs don't have to manually edit their django
settings. This hack simply adds `simmate.website.test_app` to our INSTALLED_APPS
setting. The test app is required for testing abstract models as it creates some
test tables for us -- ones that we don't want outside of pytest.
"""

# I don't want to list out all the settings used in the other file, but still
# need all of them imported an environment variable here. So I use this
# shortcut
from .settings import *

# of all the settings we just imported, we only need to update the installed apps
INSTALLED_APPS.append("simmate.website.test_app.apps.TestAppConfig")



# -*- coding: utf-8 -*-

import os

from simmate.calculators.vasp.inputs import Incar


def test_incar(tmpdir, structure):

    # TODO: make a fixture of INCAR settings to iterate through.

    # These settings are taken from the relaxation/quality00 task
    settings1 = dict(
        PREC="Low",
        EDIFF=2e-3,
        EDIFFG=-2e-1,
        ISIF=4,
        NSW=75,
        IBRION=2,
        POTIM=0.02,
        LCHARG=False,
        LWAVE=False,
        KSPACING=0.75,
        multiple_keywords__smart_ismear={
            "metal": dict(
                ISMEAR=1,
                SIGMA=0.1,
            ),
            "non-metal": dict(
                ISMEAR=0,
                SIGMA=0.05,
            ),
        },
    )

    # These settings are taken from the relaxation/mat-proj task
    settings2 = dict(
        ALGO="Fast",
        EDIFF__per_atom=5.0e-05,
        ENCUT=520,
        IBRION=2,
        ISIF=3,
        ISMEAR=-5,
        ISPIN=2,
        ISYM=0,
        KSPACING=0.4,
        LASPH=True,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        NELM=100,
        NSW=99,
        PREC="Accurate",
        SIGMA=0.05,
        MAGMOM__smart_magmom={
            "default": 0.6,
            "Ce": 5,
            "Ce3+": 1,
            "Co": 0.6,
            "Co3+": 0.6,
            "Co4+": 1,
            "Cr": 5,
            "Dy3+": 5,
            "Er3+": 3,
            "Eu": 10,
            "Eu2+": 7,
            "Eu3+": 6,
            "Fe": 5,
            "Gd3+": 7,
            "Ho3+": 4,
            "La3+": 0.6,
            "Lu3+": 0.6,
            "Mn": 5,
            "Mn3+": 4,
            "Mn4+": 3,
            "Mo": 5,
            "Nd3+": 3,
            "Ni": 5,
            "Pm3+": 4,
            "Pr3+": 2,
            "Sm3+": 5,
            "Tb3+": 6,
            "Tm3+": 2,
            "V": 5,
            "W": 5,
            "Yb3+": 1,
        },
        multiple_keywords__smart_ldau=dict(
            LDAU__auto=True,
            LDAUTYPE=2,
            LDAUPRINT=1,
            LDAUJ={},
            LDAUL={
                "F": {
                    "Co": 2,
                    "Cr": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Ni": 2,
                    "V": 2,
                    "W": 2,
                },
                "O": {
                    "Co": 2,
                    "Cr": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Ni": 2,
                    "V": 2,
                    "W": 2,
                },
            },
            LDAUU={
                "F": {
                    "Co": 3.32,
                    "Cr": 3.7,
                    "Fe": 5.3,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Ni": 6.2,
                    "V": 3.25,
                    "W": 6.2,
                },
                "O": {
                    "Co": 3.32,
                    "Cr": 3.7,
                    "Fe": 5.3,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Ni": 6.2,
                    "V": 3.25,
                    "W": 6.2,
                },
            },
        ),
    )

    # These settings are taken from the relaxation/mit task
    settings3 = dict(
        ALGO="Fast",
        EDIFF=1.0e-05,
        ENCUT=520,
        IBRION=2,
        ICHARG=1,
        ISIF=3,
        ISPIN=2,
        ISYM=0,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        NELM=200,
        NELMIN=6,
        NSW=99,
        PREC="Accurate",
        KSPACING=0.5,
        MAGMOM__smart_magmom={
            "default": 0.6,
            "Ce": 5,
            "Ce3+": 1,
            "Co": 0.6,
            "Co3+": 0.6,
            "Co4+": 1,
            "Cr": 5,
            "Dy3+": 5,
            "Er3+": 3,
            "Eu": 10,
            "Eu2+": 7,
            "Eu3+": 6,
            "Fe": 5,
            "Gd3+": 7,
            "Ho3+": 4,
            "La3+": 0.6,
            "Lu3+": 0.6,
            "Mn": 5,
            "Mn3+": 4,
            "Mn4+": 3,
            "Mo": 5,
            "Nd3+": 3,
            "Ni": 5,
            "Pm3+": 4,
            "Pr3+": 2,
            "Sm3+": 5,
            "Tb3+": 6,
            "Tm3+": 2,
            "V": 5,
            "W": 5,
            "Yb3+": 1,
        },
        multiple_keywords__smart_ismear={
            "metal": dict(
                ISMEAR=2,
                SIGMA=0.2,
            ),
            "non-metal": dict(
                ISMEAR=-5,
                SIGMA=0.05,
            ),
        },
        multiple_keywords__smart_ldau=dict(
            LDAU__auto=True,
            LDAUTYPE=2,
            LDAUPRINT=1,
            LDAUJ={},
            LDAUL={
                "F": {
                    "Ag": 2,
                    "Co": 2,
                    "Cr": 2,
                    "Cu": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Nb": 2,
                    "Ni": 2,
                    "Re": 2,
                    "Ta": 2,
                    "V": 2,
                    "W": 2,
                },
                "O": {
                    "Ag": 2,
                    "Co": 2,
                    "Cr": 2,
                    "Cu": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Nb": 2,
                    "Ni": 2,
                    "Re": 2,
                    "Ta": 2,
                    "V": 2,
                    "W": 2,
                },
                "S": {
                    "Fe": 2,
                    "Mn": 2.5,
                },
            },
            LDAUU={
                "F": {
                    "Ag": 1.5,
                    "Co": 3.4,
                    "Cr": 3.5,
                    "Cu": 4,
                    "Fe": 4.0,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Nb": 1.5,
                    "Ni": 6,
                    "Re": 2,
                    "Ta": 2,
                    "V": 3.1,
                    "W": 4.0,
                },
                "O": {
                    "Ag": 1.5,
                    "Co": 3.4,
                    "Cr": 3.5,
                    "Cu": 4,
                    "Fe": 4.0,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Nb": 1.5,
                    "Ni": 6,
                    "Re": 2,
                    "Ta": 2,
                    "V": 3.1,
                    "W": 4.0,
                },
                "S": {
                    "Fe": 1.9,
                    "Mn": 2.5,
                },
            },
        ),
    )

    incar1 = Incar(**settings1)
    incar1.to_file(
        filename=os.path.join(tmpdir, "INCAR"),
        structure=structure,
    )

    incar2 = Incar(**settings2)
    incar2.to_file(
        filename=os.path.join(tmpdir, "INCAR"),
        structure=structure,
    )

    incar3 = Incar(**settings3)
    incar3.to_file(
        filename=os.path.join(tmpdir, "INCAR"),
        structure=structure,
    )

    # simply check that we can reload. Checking for values and logic should be
    # done with individual keyword modifiers
    incar4 = Incar.from_file(os.path.join(tmpdir, "INCAR"))

    # TODO: assert equal to ....
    diff = incar1.compare_incars(incar2)


def test_custom_keyword_modifier(tmpdir, sample_structures):

    structure = sample_structures["C_mp-48_primitive"]

    incar_filename = os.path.join(tmpdir, "INCAR")

    # make a dummy modifier that just multiplies input by 2
    def keyword_modifier_dummy(structure, value):
        return value * 2

    Incar.add_keyword_modifier(keyword_modifier_dummy)

    # simple tests to confirm method is registered
    assert hasattr(Incar, "keyword_modifier_dummy")
    assert Incar.keyword_modifier_dummy(None, 4) == 8

    # now test out logic with keyword
    incar = Incar(ENCUT__dummy=4)
    incar.to_file(
        filename=incar_filename,
        structure=structure,
    )

    incar2 = Incar.from_file(incar_filename)
    assert incar2.get("ENCUT", None) == 8



# -*- coding: utf-8 -*-

"""

> :warning: This file is only for use by the Simmate team. Users should instead
access data via the load_remote_archive method.

This file is for pulling OQMD data into the Simmate database. 

There are many ways to pull from this database, but it looks like the easiest
is the qmpy_rester python package. This officially supported and maintained
at https://github.com/mohanliu/qmpy_rester. For now the package is only available
via a pip install.

For other options such as the REST API, check out http://oqmd.org/static/docs/restful.html

"""

import os

from tqdm import tqdm
import pandas
from django.db import transaction

from simmate.toolkit import Structure
from simmate.database.third_parties import OqmdStructure


@transaction.atomic
def load_all_structures_from_api():
    """
    Only use this function if you are part of the Simmate dev team!

    Loads all structures directly for the OQMD database into the local
    Simmate database.
    """

    # OQMD is not a dependency of simmate, so make sure you install it before using
    # this module
    try:
        import qmpy_rester
    except:
        raise ModuleNotFoundError(
            "You must install qmpy-rester with `pip install qmpy_rester`"
        )
    # The documentation indicates that we handle a query via a context manager
    # for qmpy_rester. Each query is returned as a page of data, where we need
    # to iterate through all of the pages. To do this, we constantly make a query
    # and check the "next" field, which tells us if its the last page or not.

    # as we download all the data, we store it in a main list
    data = []

    # starting from the first page, assume we aren't on the last page until told
    # otherwise. And loop until we know its the last page.
    current_page = 0
    is_last_page = False
    results_per_page = 100  # based on OQMD recommendations
    while not is_last_page:

        with qmpy_rester.QMPYRester() as query:

            # make the query
            result = query.get_oqmd_phases(
                verbose=False,
                limit=results_per_page,
                offset=current_page * results_per_page,
                #
                # Note delta_e is the formation energy and then stability is the
                # energy above hull.
                fields="entry_id,unit_cell,sites,delta_e",
                # element_set="Al,C",  # Useful for testing
            )
            # grab the data for the next slice of structures
            query_slice = result["data"]
            # And check to see if this is the last page. The logic here is if
            # there is data for "next", then it isn't the last page
            is_last_page = not bool(result["links"]["next"])

            # store the slice of entrys in our main list
            for entry in query_slice:
                data.append(entry)
            # on the very first page, let's also check the total number of pages
            if current_page == 0:
                total_pages = result["meta"]["data_available"] // results_per_page
            # move on to the next page
            print(f"Successfully downloaded page {current_page} of {total_pages}")
            current_page += 1
    # Now iterate through all the data -- which is a list of dictionaries.
    # We convert the data into a pymatgen object and sanitize it before saving
    # to the Simmate database
    for entry in tqdm(data):

        # Parse the data into a pymatgen object
        # Also before converting into a pymatgen object, we need to parse the sites,
        # which are given as a list of "Element @ X Y Z" (ex: "Na @ 0.5 0.5 0.5")
        # Changing this format is why we have this complex lists below
        structure = Structure(
            lattice=entry["unit_cell"],
            species=[site.split(" @ ")[0] for site in entry["sites"]],
            coords=[
                [float(n) for n in site.split(" @ ")[1].split()]
                for site in entry["sites"]
            ],
            coords_are_cartesian=False,
        )

        # now convert the entry to a database object
        structure_db = OqmdStructure.from_toolkit(
            id="oqmd-" + str(entry["entry_id"]),
            structure=structure,
            energy=entry["delta_e"],
        )

        # and save it to our database!
        structure_db.save()


@transaction.atomic
def load_all_structures_from_files(
    base_directory: str = "oqmd",
    only_add_new_cifs: bool = True,
):
    """
    Jiahong Shen was kind enough to provide all the crystal structures from
    the OQMD as POSCAR files. This makes loading the structures into the
    Simmate database much faster as we are no longer bottlenecked by the REST
    API and internet connections.

    All POSCARs are in the same folder, where the name of each is the
    <id>-<composition> (ex: 12345-NaCl). There are also csv's that contain
    additional data such as the energy:

        - all_oqmd_entry.csv
        - all_public_entries.csv
        - all_public_fes.csv
        - get_all_entry_id_public.py
        - get_all_entry_poscar.py

    There are currently 1,013,654 structures and this function takes roughly
    3hrs to run.
    """

    # load the csv that contains the list of filenames and their values
    df = pandas.read_csv(os.path.join(base_directory, "all_oqmd_entry.csv"))

    # iterate through the list and load the structures to our database!
    # Use tqdm to monitor progress.
    for _, row in tqdm(df.iterrows()):

        # load the structure from the poscar file
        filename = os.path.join(base_directory, row.filename)
        with open(filename) as file:
            contents = file.read()
        structure = Structure.from_str(contents, "poscar")

        # save the data to the Simmate database
        # now convert the entry to a database object
        try:
            structure_db = OqmdStructure.from_toolkit(
                id="oqmd-" + str(row.entry_id),
                structure=structure,
                formation_energy=row.formationenergy,
            )
            structure_db.save()
        # A few structures fail because of the symmetry analyzer can't determine
        # the spacegroup. These are...
        # 1443135, 1443014, 1451986, 1452015, 1452024
        except:
            print(f"Failed: {row.entry_id}")



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod

from dask.diagnostics import ProgressBar
from dask import bag


class Validator(ABC):
    @abstractmethod
    def check_structure(self, structure):
        # User needs to define this function where one structure is input.
        # The output should be a True or False value that indicates whether
        # or not the structure passed the check.
        raise NotImplementedError

    def check_many_structures(self, structures, progressbar=True, mode="threads"):
        # USING DASK TO PARALLELIZE
        structures_bag = bag.from_sequence(structures)
        validation_bag = structures_bag.map(self.check_structure)

        # Based on whether the user wants a progress bar or not, decide how
        # the compute() command is called.
        if progressbar:
            with ProgressBar():
                checks = validation_bag.compute(scheduler=mode)
        else:
            checks = validation_bag.compute(scheduler=mode)

        return checks



# -*- coding: utf-8 -*-

import os
from typing import Tuple

from simmate.toolkit import Structure
from simmate.toolkit.diffusion import MigrationHop

# TODO: use MigrationImages use to generate images?

from simmate.workflow_engine import Workflow, task
from simmate.workflow_engine.common_tasks import (
    load_input_and_register,
    parse_multi_command,
)

from simmate.calculators.vasp.workflows.diffusion.utilities import (
    get_migration_images_from_endpoints,
)
from simmate.calculators.vasp.workflows.relaxation.neb_endpoint import (
    Relaxation__Vasp__NebEndpoint,
)
from simmate.calculators.vasp.workflows.diffusion.neb_from_images import (
    Diffusion__Vasp__NebFromImages,
)
from simmate.calculators.vasp.database.nudged_elastic_band import (
    MITDiffusionAnalysis,
)


class Diffusion__Vasp__NebSinglePath(Workflow):
    """
    Runs a full diffusion analysis on a single migration hop using NEB.

    Typically, this workflow is not ran directly -- but instead, you would
    use diffusion/all-paths which submits a series of this workflow for you.

    For a given migration path, supercell structures are generated for
    the start and end points of the migration, where vacancy-based diffsion
    is used. These supercells are relaxed using the relaxation/neb-endpoint
    workflow and then interpolated to generate 7 images. These 7 images are
    then relaxed using NEB within the diffusion/from-images workflow.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation/neb-endpoint (for both start and end supercells)
        - a mini task that interpolated relaxed endpoints and makes images
        - diffusion/from-images

    Currently, this workflow will not run through the command-line, as we
    have not implemented a file format for MigrationHop's yet.
    """

    database_table = MITDiffusionAnalysis
    description_doc_short = "runs NEB using a MigrationHop object as input"

    @classmethod
    def run_config(
        cls,
        migration_hop: MigrationHop,
        directory: str = None,
        source: dict = None,
        command: str = None,
        # These help link results to a higher-level table.
        diffusion_analysis_id: int = None,
        migration_hop_id: int = None,
        # TODO: Can the hop id be inferred from the migration_hop or somewhere
        # else in this context? Maybe even load_input_and_register will use
        # prefect id once it's a Calculation?
        is_restart: bool = False,
    ):

        # split the command if separate ones were given
        subcommands = parse_multi_command(
            command,
            commands_out=["command_supercell", "command_neb"],
        )

        # Load our input and make a base directory for all other workflows to run
        # within for us.
        parameters_cleaned = load_input_and_register(
            migration_hop=migration_hop,
            source=source,
            directory=directory,
            command=command,
            diffusion_analysis_id=diffusion_analysis_id,
            migration_hop_id=migration_hop_id,
            is_restart=is_restart,
            register_run=False,
        )

        # get the supercell endpoint structures
        supercell_start, supercell_end = get_endpoint_structures(
            parameters_cleaned["migration_hop"]
        )

        # Relax the starting supercell structure
        endpoint_start_state = Relaxation__Vasp__NebEndpoint.run(
            structure=supercell_start,
            command=subcommands["command_supercell"],
            directory=parameters_cleaned["directory"]
            + os.path.sep
            + f"{Relaxation__Vasp__NebEndpoint.name_full}.start",
            is_restart=is_restart,
        )

        # Relax the ending supercell structure
        endpoint_end_state = Relaxation__Vasp__NebEndpoint.run(
            structure=supercell_end,
            command=subcommands["command_supercell"],
            directory=parameters_cleaned["directory"]
            + os.path.sep
            + f"{Relaxation__Vasp__NebEndpoint.name_full}.end",
            is_restart=is_restart,
        )

        # wait for the endpoint relaxations to finish
        endpoint_start_result = endpoint_start_state.result()
        endpoint_end_result = endpoint_end_state.result()

        images = get_migration_images_from_endpoints(
            supercell_start={
                "database_table": Relaxation__Vasp__NebEndpoint.database_table.__name__,
                "directory": endpoint_start_result["directory"],
                "structure_field": "structure_final",
            },
            supercell_end={
                "database_table": Relaxation__Vasp__NebEndpoint.database_table.__name__,
                "directory": endpoint_end_result["directory"],
                "structure_field": "structure_final",
            },
        )

        # Run NEB on this set of images
        neb_state = Diffusion__Vasp__NebFromImages.run(
            migration_images=images,
            command=subcommands["command_neb"],
            source=parameters_cleaned["source"],
            directory=parameters_cleaned["directory"],
            diffusion_analysis_id=diffusion_analysis_id,
            migration_hop_id=migration_hop_id,
            is_restart=is_restart,
        )


@task
def get_endpoint_structures(migration_hop: MigrationHop) -> Tuple[Structure]:
    """
    Simple wrapper for get_sc_structures method that makes it a Prefect task.
    I assume parameters for now
    """
    start_supercell, end_supercell, _ = migration_hop.get_sc_structures(
        # min_atoms=10,
        # max_atoms=50,
        # min_length=5,
        vac_mode=True,
    )
    try:
        assert start_supercell != end_supercell
    except:
        raise Exception(
            "This structure has a bug due to a rounding error. "
            "Our team is aware of this bug and it has been fixed for the next "
            "pymatgen-analysis-diffusion release."
        )
    return start_supercell, end_supercell



# -*- coding: utf-8 -*-

from .async_wrapper import async_to_sync
from .files import (
    get_directory,
    make_archive,
    make_error_archive,
    archive_old_runs,
    empty_directory,
    copy_directory,
)
from .other import (
    get_conda_env,
    get_doc_from_readme,
    get_chemical_subsystems,
    get_latest_version,
    check_if_using_latest_version,
)

# must be last to prevent circular import
__doc__ = get_doc_from_readme(__file__)



# -*- coding: utf-8 -*-

from typing import List, Callable
from dask.distributed import wait, TimeoutError
from simmate.configuration.dask.client import get_dask_client


def batch_submit(
    function: Callable,
    args_list: List[dict],
    batch_size: int,
    batch_timeout: float = None,
):
    """
    Given a function and a list of inputs that should be iterated over, this
    submits all inputs to a Dask local cluster in batches.

    This function has very specific use-cases, such as when we are submitting
    >100,000 tasks and each task is unstable / writing to the database. Therefore,
    you should test out Dask normally before trying this utility. Always give
    preference to Dask's `client.map` method over this utility.

    #### Parameters

    - `function`:
        Function that each kwargs entry should be called with.
    - `args_list`:
        A list of parameters that will each be submitted to function via
        function(*args).
    - `batch_size`:
        The number of calls to submit at a time. No new jobs will be
        submitted until the entire preceeding batch completes.
    - `batch_timeout`:
        The timelimit to wait for any given batch before cancelling the remaining
        runs. No error will be raised when jobs are cancelled. The default is
        no timelimit.
    """
    # TODO: I'd like to support a list of kwargs as well
    # TODO: should I add an option to return the results?

    # grab the Dask client
    client = get_dask_client()

    # Iterate through our inputs and submit them to the Dask cluster in batches
    for i in range(0, len(args_list), batch_size):
        chunk = args_list[i : i + batch_size]
        futures = client.map(
            function,
            chunk,
            pure=False,
        )
        try:
            wait(futures, timeout=batch_timeout)
        except TimeoutError:
            client.cancel(futures)



# -*- coding: utf-8 -*-

from django.urls import path

from simmate.website.core_components import views

urlpatterns = [
    # This view is meant to be used as an iFrame for viewing 3D crystal structures
    path(
        route="structure-viewer/",
        view=views.structure_viewer,
        name="structure_viewer",
    ),
    #
    # This two views are for the Spacegroup table. "symmetry" lists all of the
    # 230 spacegroups while "spacegroup" is for a single one.
    path(
        route="symmetry/",
        view=views.SymmetryViewSet.list_view,
        name="symmetry",
    ),
    path(
        route="symmetry/<pk>/",
        view=views.SymmetryViewSet.retrieve_view,
        name="spacegroup",
    ),
    #
    # This view is strictly for testing different components and making sure
    # they are working.
    path(
        route="test/",
        view=views.test_viewer,
        name="test_viewer",
    ),
]



# -*- coding: utf-8 -*-

import os

from simmate.conftest import copy_test_files, make_dummy_files
from simmate.calculators.vasp.inputs import Potcar
from simmate.calculators.vasp.tasks.band_structure import MatprojBandStructure
from simmate.calculators.vasp.inputs.potcar_mappings import PBE_ELEMENT_MAPPINGS


def test_band_structure_setup(structure, tmpdir, mocker):

    # estabilish filenames that we make and commonly reference
    incar_filename = os.path.join(tmpdir, "INCAR")
    poscar_filename = os.path.join(tmpdir, "POSCAR")
    potcar_filename = os.path.join(tmpdir, "POTCAR")

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # try to make input files in the tmpdir
    MatprojBandStructure.setup(directory=tmpdir, structure=structure)
    assert os.path.exists(incar_filename)
    assert os.path.exists(poscar_filename)
    assert os.path.exists(potcar_filename)
    Potcar.to_file_from_type.assert_called_with(
        structure.composition.elements,
        "PBE",
        potcar_filename,
        PBE_ELEMENT_MAPPINGS,
    )


def test_band_structure_workup(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="band_structure",
    )

    # init main task
    task = MatprojBandStructure()

    # estabilish filenames that we make and commonly reference
    summary_filename = os.path.join(tmpdir, "simmate_summary.yaml")
    plot_filename = os.path.join(tmpdir, "band_structure.png")

    # run the full workup
    task.workup(tmpdir)
    assert os.path.exists(summary_filename)
    assert os.path.exists(plot_filename)



# -*- coding: utf-8 -*-

from .neb_all_paths import Diffusion__Vasp__NebAllPaths
from .neb_single_path import Diffusion__Vasp__NebSinglePath
from .neb_from_endpoints import Diffusion__Vasp__NebFromEndpoints
from .neb_from_images import Diffusion__Vasp__NebFromImages



# -*- coding: utf-8 -*-

# Example job(s) [with and without subprocess.Popen]
# Example handler(s)

# Test...
# setup a working directory when given None, tempdir, and string of dir
# non-zero exit code raises error
# output compression (assert location and content)
# log output (assert naming, content, and location)
# return corrections (on and off)
# a simple SupervisedJobTask run (without handlers, with, and with monitors)
# run with and without handlers
# run with and without monitors
# terminate with a monitor
# terminate with a special-case monitor
# catch error with a non-monitor
# test max_errors limit

import os

import pytest

from prefect import flow

from simmate.workflow_engine import ErrorHandler, S3Task
from simmate.workflow_engine.supervised_staged_shell_task import (
    NonZeroExitError,
    MaxCorrectionsError,
)

# ----------------------------------------------------------------------------

# make some a simple ErrorHandlers for us to test with


class AlwaysPassesHandler(ErrorHandler):
    def check(self, directory):
        return False

    def correct(self, directory):
        # this should never be entered since check() never returns an error
        raise Exception


class AlwaysFailsHandler(ErrorHandler):
    def check(self, directory):
        return True

    def correct(self, directory):
        return "ExampleCorrection"


class AlwaysPassesMonitor(AlwaysPassesHandler):
    is_monitor = True


class AlwaysFailsMonitor(AlwaysFailsHandler):
    is_monitor = True


class AlwaysPassesSpecialMonitor(AlwaysPassesMonitor):
    has_custom_termination = True

    def terminate_job(self, directory, **kwargs):
        return True


class AlwaysFailsSpecialMonitorRetry(AlwaysFailsMonitor):
    has_custom_termination = True

    def terminate_job(self, directory, **kwargs):
        return True


class AlwaysFailsSpecialMonitorNoRetry(AlwaysFailsMonitor):
    has_custom_termination = True

    def terminate_job(self, directory, **kwargs):
        return False


# ----------------------------------------------------------------------------


@pytest.mark.prefect_db
def test_s3task_methods():
    class DummyTask(S3Task):
        command = "echo dummy"

    assert isinstance(DummyTask.get_config(), dict)

    DummyTask.print_config()  # a print statment w. nothing else to check

    DummyTask.to_prefect_task()  # unused for now

    @flow
    def test():
        task = DummyTask.run()
        return task.result()

    state = test(return_state=True)
    result = state.result()

    assert state.is_completed()
    assert os.path.exists(result["directory"])
    os.rmdir(result["directory"])


def test_s3task_1():
    # run a basic task w.o. any handlers or monitoring

    class DummyTask(S3Task):
        command = "echo dummy"
        monitor = False

    output = DummyTask.run_config()

    assert output["result"] == None
    assert output["corrections"] == []
    assert output["prefect_flow_run_id"] == None

    # make sure that a "simmate-task-*" directory was created
    assert os.path.exists(output["directory"])

    # and delete that directory
    os.rmdir(output["directory"])


def test_s3task_2():
    # test file compression

    class DummyTask(S3Task):
        command = "echo dummy"
        monitor = False
        compress_output = True

    output = DummyTask.run_config()

    # make sure that a "simmate-task-*.zip" archive was created
    assert os.path.exists(output["directory"] + ".zip")

    # make sure that a "simmate-task-*" directory was removed
    assert not os.path.exists(output["directory"])

    # and delete the archive
    os.remove(output["directory"] + ".zip")


def test_s3task_3(tmpdir):
    # Make a task with error handlers, monitoring, and specific directory

    class DummyTask(S3Task):
        command = "echo dummy"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [
            AlwaysPassesHandler(),
            AlwaysPassesMonitor(),
            AlwaysPassesSpecialMonitor(),
        ]

    # use the temporary directory
    assert DummyTask.run_config(directory=tmpdir) == {
        "result": None,
        "corrections": [],
        "directory": tmpdir,
        "prefect_flow_run_id": None,
    }


def test_s3task_4(tmpdir):
    # test nonzero returncode

    class DummyTask(S3Task):
        command = "NonexistantCommand 404"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysPassesHandler()]

    pytest.raises(
        NonZeroExitError,
        DummyTask.run_config,
        directory=tmpdir,
    )


def test_s3task_5(tmpdir):
    # testing handler-triggered failures

    class DummyTask(S3Task):
        command = "echo dummy"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysFailsHandler()]

    pytest.raises(
        MaxCorrectionsError,
        DummyTask.run_config,
        directory=tmpdir,
    )


def test_s3task_6(tmpdir):
    # monitor failure

    class DummyTask(S3Task):
        command = "echo dummy"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysFailsMonitor()]

    pytest.raises(
        MaxCorrectionsError,
        DummyTask.run_config,
        directory=tmpdir,
    )


def test_s3task_7(tmpdir):
    # special-monitor failure (non-terminating monitor)

    class DummyTask(S3Task):
        command = "echo dummy"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysFailsSpecialMonitorRetry()]

    pytest.raises(
        MaxCorrectionsError,
        DummyTask.run_config,
        directory=tmpdir,
    )


def test_s3task_8(tmpdir):
    # check that monitor exits cleanly when retries are not allowed and no
    # workup method raises an error

    class DummyTask(S3Task):
        command = "echo dummy"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysFailsSpecialMonitorRetry()]

    result = DummyTask.run_config(directory=tmpdir)
    assert len(result["corrections"]) == 1


def test_s3task_8(tmpdir):
    # make sure an error is raised when a file is missing

    class DummyTask(S3Task):
        command = "echo dummy"
        required_files = ["FILE_THAT_DOESNT_EXIST"]

    pytest.raises(
        Exception,
        DummyTask.run_config,
        directory=tmpdir,
    )



# -*- coding: utf-8 -*-

from .matproj import StaticEnergy__Vasp__Matproj
from .mit import StaticEnergy__Vasp__Mit
from .quality04 import StaticEnergy__Vasp__Quality04
from .neb_endpoint import StaticEnergy__Vasp__NebEndpoint
